unit edgemask;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Buttons, StdCtrls, ComCtrls,  define_types,Texture_3D_Unit,OpenGL1x,
  ExtCtrls,nii_save;

type
  TEdgeForm = class(TForm)
    EdgeOnBtn: TSpeedButton;
    EdgePanel: TPanel;
    BiasGainImage: TImage;
    Label2: TLabel;
    BiasTrack: TTrackBar;
    Label3: TLabel;
    GainTrack: TTrackBar;
    EdgeOffBtn: TSpeedButton;
    procedure BiasGainTrackChange(Sender: TObject);
    procedure ToggleEdgeMap(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;
    //function Smooth3DNII (var lTexture: TTexture): integer;//returns time to complete in msec, -1 for error
 // function GaussEdge (var lTexture: TTexture; lFWHM: single): integer;
    function CubicEdge (var lTexture: TTexture; lSmoothingPasses: integer): integer;
var
  EdgeForm: TEdgeForm;

implementation

uses mainunit;

{$R *.dfm}
procedure RescaleToByte (lFBuff: singlep0; lN: integer);
//scales values 0..255
var
   lMin,lMax,lScale: single;
   lPos: integer;
begin
     if lN < 1 then
        exit;
     lMin := lFBuff^[0];
     lMax := lMin;
     for lPos := 0 to lN-1 do begin
         if lFBuff^[lPos]> lMax then
            lMax := lFBuff^[lPos];
         if lFBuff^[lPos]< lMin then
            lMin := lFBuff^[lPos];
     end;
     if lMax = lMin then
        lScale := 1
     else
         lScale := 255/(lMax-lMin);
     for lPos := 0 to lN-1 do
         lFBuff^[lPos] := ((lFBuff^[lPos]-lMin)*lScale);
end;

function Img2SingleArray (var lTexture: TTexture; var lOutImg: singlep0): integer;
var
  lVox,lX: integer;
begin
  result := lTexture.Dim[1]*lTexture.Dim[2]*lTexture.Dim[3];
  if result < 1 then
    exit;
  getmem(lOutImg,result * sizeof(single));
  if lTexture.RawUnscaledImg16 <> nil then begin //16
    for lVox := 0 to result-1 do
            lOutImg^[lVox] := lTexture.RawUnscaledImg16^[lVox];
  end else if lTexture.RawUnscaledImg32 <> nil then begin //32-bit
    for lVox := 0 to result-1 do
      lOutImg^[lVox] := lTexture.RawUnscaledImg32^[lVox];
  end else if (lTexture.RawUnscaledImg8 <> nil) then begin //8-bit
    for lVox := 0 to result-1 do
      lOutImg^[lVox] :=  lTexture.RawUnscaledImg8^[lVox];
  end else if (lTexture.RawUnscaledImgRGBA <> nil) then begin
   lX := 1; //indexed from 0, so 1 = green in RGBA quad
    for lVox := 0 to result-1 do begin
      lOutImg^[lVox] := lTexture.RawUnscaledImgRGBA^[lX];
      lX := lX + 4;
    end;
  end else begin
      Showmessage('Serious error with Edge filter: unknown datatype.');
      exit;
  end;
end;

function CubicEdge (var lTexture: TTexture; lSmoothingPasses: integer): integer;
const
  kCen : single = 2/3; //center of kernal is 2/3
  kEdge : single = 1/6; //voxel on each edge is 1/6
var
   lSrcImg,lSmoothImg,lSmoothImg2: SingleP0;
   lStartTime: DWord;
   lnPass,lPass,lSliceSz,lZPos,lYPos,lX,lY,lZ,lXDim,lYDim,lZDim,lnVox,lVox: integer;
begin
   result := -1;
   lStartTime:= GetTickCount;
   lXDim := lTexture.Dim[1];
   lYDim := lTexture.Dim[2];
   lZDim := lTexture.Dim[3];
   lSliceSz := lTexture.Dim[1]*lTexture.Dim[2];
   lnVox := lSliceSz*lTexture.Dim[3];
   lnPass := lSmoothingPasses + 1;//1 pass for edges, plus additional to smooth results
   if lnPass < 1 then //just in case someone requests a NEGATIVE number of smoothing passes
    lnPass := 1;
    if (lnVox < 0) or (lTexture.Dim[1] < 3) or (lTexture.Dim[2] < 3) or (lTexture.Dim[3] < 3) then begin
	    ShowMessage('Smooth3DNII error: Image dimensions are not large enough to filter.');
	    exit;
   end;
   if Img2SingleArray (lTexture,lSrcImg) < 1 then
    exit;
  getmem(lSmoothImg,lnVox*sizeof(Single));
  getmem(lSmoothImg2,lnVox*sizeof(Single));
  Move (lSrcImg^,lSmoothImg^, lnVox *  sizeof(single));
  Move (lSrcImg^,lSmoothImg2^, lnVox *  sizeof(single));
 for lPass := 1 to lnPass do begin //first pass make edges, second pass smooth edges....
  //X-direction - copy from SmoothImg -> SmoothImg2
  for lZ := 2 to lZdim-1 do begin
    lZPos := (lZ-1)*lSliceSz;
    for lY := 2 to lYdim-1 do begin
      lYPos := (lY-1)*lXdim;
      for lX := 2 to lXdim-1 do begin
            lVox := lZPos+lYPos+lX-1;//-1 as indexed from 0
            lSmoothImg2^[lVox] := (lSmoothImg^[lVox-1]*kEdge)+(lSmoothImg^[lVox]*kCen)+(lSmoothImg^[lVox+1]*kEdge);
      end; {lX}
    end; {lY}
  end; {lZ loop for X-plane}
  //Y-direction - copy from SmoothImg2 -> SmoothImg
  for lZ := 2 to lZdim-1 do begin
    lZPos := (lZ-1)*lSliceSz;
    for lY := 2 to lYdim-1 do begin
      lYPos := (lY-1)*lXdim;
      for lX := 2 to lXdim-1 do begin
            lVox := lZPos+lYPos+lX-1;//-1 as indexed from 0
            lSmoothImg^[lVox] := (lSmoothImg2^[lVox-lXdim]*kEdge)+(lSmoothImg2^[lVox]*kCen)+(lSmoothImg2^[lVox+lXdim]*kEdge);
      end; {lX}
    end; {lY}
  end; {lZ loop for Y-plane}
  //Z-direction - copy from SmoothImg -> SmoothImg2
  for lZ := 2 to lZdim-1 do begin
    lZPos := (lZ-1)*lSliceSz;
    for lY := 2 to lYdim-1 do begin
      lYPos := (lY-1)*lXdim;
      for lX := 2 to lXdim-1 do begin
            lVox := lZPos+lYPos+lX-1;//-1 as indexed from 0
            lSmoothImg2^[lVox] := (lSmoothImg^[lVox-lSliceSz]*kEdge)+(lSmoothImg^[lVox]*kCen)+(lSmoothImg^[lVox+lSliceSz]*kEdge);
      end; {lX}
    end; {lY}
  end; {lZ loop for X-plane}
   if lPass = 1 then begin//create a difference map
     for lVox := 0 to lnVox-1 do
       lSmoothImg2^[lVox] :=  abs(lSmoothImg2^[lVox]-lSrcImg^[lVox]);
     if lPass <> lnPass then //make this the new source image
      Move (lSmoothImg2^,lSmoothImg^, lnVox *  sizeof(single));
   end;
  end; //for each pass...
   //next make this in the range 0..255
    RescaleToByte (lSmoothImg2,lnVox);
    SetLengthB(lTexture.EdgeImg,lnVox);
    for lVox := 0 to lnVox-1 do
       lTexture.EdgeImg^[lVox] :=  round(lSmoothImg2^[lVox]);
    //   lTexture.RawUnscaledImg8^[lVox] :=  round(lSmoothImg2^[lVox]);
     Freemem(lSmoothImg);
     Freemem(lSmoothImg2);
     Freemem(lSrcImg);
     result := GetTickCount-lStartTime; //point of no return - after this the user will need to freemem
     GLForm1.MIPChange(nil);
end;

(* Actual gaussian smooth is pretty slow
type
  TKernel =  RECORD //peristimulus plot
    Width: integer; //e.g. with a width of 2 the filter will go from -2..2
    CutWt,Ht: array of single;
  end;

function CreateKernel(var lK: TKernel; FWHMpix: single):integer;
//returns full width of kernel.. -n..+n : image should be at least this wide
//FWHM in pixels, (e.g.   lFWHM / Pixmm)
var
  lI2,lI: integer;
  lCumGauss,lexpd,lsig: double;
begin
   result := maxint;
	 lsig  := FWHMpix/sqrt(8*ln(2));  //FWHM -> sigma
   if lsig = 0 then
    exit;
	 lK.Width  := round(6*lsig);
   setlength(lK.CutWt,lK.Width+1);
   setlength(lK.Ht,lK.Width+1);
   lexpd := 2*lsig*lsig;
   for lI := 0 to lK.Width do 
			lK.Ht[lI] := exp(-1*(lI*lI)/lexpd) ;
   lCumGauss := 0;
   for lI := -lK.Width to lK.Width do
			lCumGauss := lCumGauss + lK.Ht[abs(lI)];
   if lCumGauss = 0 then
    exit;
    for lI := 0 to lK.Width do
      lK.Ht[lI] := lK.Ht[lI]/lCumGauss;
   for lI := 0 to lK.Width do begin
      lCumGauss := 0;
			for lI2 := (lK.Width - lI) downto -lK.Width  do
				lCumGauss := lCumGauss + lK.Ht[abs(lI2)];
			//if lCumGauss <> 0 then
      lK.CutWt[lI] := lCumGauss;
  end;
  result := lK.Width*lK.Width;
end;


procedure FreeKernel(var lK: TKernel);
begin
  lK.Ht := nil;
  lK.CutWt := nil;
end;

procedure Bound( var lLo,lHi,lClipped: integer; lMin,lMax: integer);
begin
  lClipped := (lHi-lLo); //range
  if lLo < lMin then
    lLo := lMin;
  if lHi > lMax then
    lHi := lMax;
  lClipped := lClipped- (lHi-lLo);
end;

function GaussEdge (var lTexture: TTexture; lFWHM: single): integer;
label
  666;
var
   lXK,lYK,lZK: TKernel;
   lSrcImg,lSmoothImg,lSmoothImg2: SingleP0;
   lStartTime: DWord;
   lSum: single;
   F,lLo,lHi,lClipped,lSliceSz,lZPos,lYPos,lX,lY,lZ,lXDim,lYDim,lZDim,lnVox,lVox: integer;
begin
   result := -1;
   lStartTime:= GetTickCount;
   lXDim := lTexture.Dim[1];
   lYDim := lTexture.Dim[2];
   lZDim := lTexture.Dim[3];
   lSliceSz := lTexture.Dim[1]*lTexture.Dim[2];
   lnVox := lSliceSz*lTexture.Dim[3];
    if (lnVox < 0) or (lTexture.Dim[1] < 3) or (lTexture.Dim[2] < 3) or (lTexture.Dim[3] < 3) then begin
	    ShowMessage('Smooth3DNII error: Image dimensions are not large enough to filter.');
	    exit;
   end;
   if  CreateKernel(lXK, lFWHM / lTexture.PixMM[1])> lXDim then
    goto 666;
   if not CreateKernel(lYK, lFWHM / lTexture.PixMM[2])> lYDim then
    goto 666;
   if not CreateKernel(lZK, lFWHM / lTexture.PixMM[3])> lZDim then
    goto 666;
   if Img2SingleArray (lTexture,lSrcImg) < 1 then
    goto 666;
  getmem(lSmoothImg,lnVox*sizeof(Single));
  getmem(lSmoothImg2,lnVox*sizeof(Single));
  Move (lSrcImg^,lSmoothImg^, lnVox *  sizeof(single));
  Move (lSrcImg^,lSmoothImg2^, lnVox *  sizeof(single));
//showmessage(inttostr(lXK.Width));
  //X-direction - copy from SmoothImg -> SmoothImg2
  for lZ := 1 to lZdim do begin
    lZPos := (lZ-1)*lSliceSz;
    for lY := 1 to lYdim do begin
      lYPos := (lY-1)*lXdim;
      for lX := 1 to lXdim do begin
        lLo := lX-lXK.Width;
        lHi := lX+lXK.Width;
        Bound( lLo,lHi,lClipped, 1,lXDim);
        lSum := 0;
        lVox := lZPos+lYPos+lX-1;
        for F := lLo to lHi do
            lSum := lSum + lXK.Ht[abs(F-lX)]*lSmoothImg^[lZPos+lYPos+F-1];
        //-1 as indexed from 0
        //fx(lLo,lHi,lClipped);
        lSmoothImg2^[lVox] := lSum/lXK.CutWt[lClipped];

      end; {lX}
    end; {lY}
  end; {lZ loop for X-plane} 
  //Y-direction - copy from SmoothImg2 -> SmoothImg
  for lZ := 1 to lZdim do begin
    lZPos := (lZ-1)*lSliceSz;
    for lY := 1 to lYdim do begin
      lYPos := (lY-1)*lXdim;
      lLo := lY-lYK.Width;
      lHi := lY+lYK.Width;
      Bound( lLo,lHi,lClipped, 1,lYDim);
      for lX := 1 to lXdim do begin
        lSum := 0;
        lVox := lZPos+lYPos+lX-1;
        for F := lLo to lHi do
            lSum := lSum + lYK.Ht[abs(F-lY)]*lSmoothImg2^[lZPos+((F-1)*lXDim)+lX-1];
        //-1 as indexed from 0
        lSmoothImg^[lVox] := lSum/lYK.CutWt[lClipped]
      end; {lX}
    end; {lY}
  end; {lZ loop for Y-plane}
  //Z-direction - copy from SmoothImg -> SmoothImg2
  for lZ := 1 to lZdim do begin
    lZPos := (lZ-1)*lSliceSz;
    lLo := lZ-lZK.Width;
    lHi := lZ+lZK.Width;
    Bound( lLo,lHi,lClipped, 1,lZDim);
    for lY := 1 to lYdim do begin
      lYPos := (lY-1)*lXdim;
      for lX := 1 to lXdim do begin
        lSum := 0;
        lVox := lZPos+lYPos+lX-1;
        for F := lLo to lHi do
            lSum := lSum + lZK.Ht[abs(F-lZ)]*lSmoothImg^[((F-1)*lSliceSz)+lYPos+lX-1];
        //-1 as indexed from 0
        lSmoothImg2^[lVox] := lSum/lZK.CutWt[lClipped]
      end; {lX}
    end; {lY}
  end; {lZ loop for X-plane}

  FreeKernel(lXK);
  FreeKernel(lYK);
  FreeKernel(lZK);
     for lVox := 0 to lnVox-1 do
       lSmoothImg2^[lVox] :=  abs(lSmoothImg2^[lVox]-lSrcImg^[lVox]);
   //next make this in the range 0..255
    RescaleToByte (lSmoothImg2,lnVox);
    SetLengthB(lTexture.EdgeImg,lnVox);
    for lVox := 0 to lnVox-1 do
       lTexture.EdgeImg^[lVox] :=  round(lSmoothImg2^[lVox]);
    //   lTexture.RawUnscaledImg8^[lVox] :=  round(lSmoothImg2^[lVox]);
     Freemem(lSmoothImg);
     Freemem(lSmoothImg2);
     Freemem(lSrcImg);
     result := GetTickCount-lStartTime; //point of no return - after this the user will need to freemem

    exit;
   666:
   showmessage('Error with Gaussian smooth: perhaps image is too small or FWHM too wide.');
end;     *)


procedure DrawBiasGain (lImage: TImage; lBias,lGain: integer);
var
  Vi,Hi,Xi,Yi: integer;
  lLUT: TLUTb ;
begin
  with lImage.Canvas do begin
    //Kernel := TKernelResampler(Src.Bitmap.Resampler).Kernel;
    Pen.Width := 1;
    Vi := lImage.Height;
    Hi := lImage.Width;
    Brush.Color := clBlack;
    Brush.Style := bsSolid;
    Pen.Color := clBlack;
    Pen.Width := 1;
    Rectangle(0,0,Hi+1,Vi+1);
    Pen.Color := clGray;
    //horizontal lines
    Pen.Color := $3400F4;//clRed;
    EdgeBiasGain (lBias,lGain,  lLUT);
    Vi := Vi-1;
    Hi := Hi-1;
    //Yi := Vi-round(lLUT[0]/255*Vi);
    for Xi := 0 to Hi do begin
        //MoveTo(Xi-1,Yi);
        MoveTo(Xi,Vi);
        Yi := Vi-round(lLUT[round(Xi/Hi*255)]/255*Vi);
        LineTo(Xi,Yi);
    end;
  end;//with image
end; //DrawBiasGain

procedure TEdgeForm.BiasGainTrackChange(Sender: TObject);
begin
   if gTexture3D.EdgeImg = nil then begin
     ToggleEdgeMap(nil);
     exit;
   end;
   gTexture3D.EdgeBias := BiasTrack.position;
   gTexture3D.EdgeGain := GainTrack.position;

    DrawBiasGain (BiasGainImage,BiasTrack.position,GainTrack.position);
    GLForm1.CLUTTimerXOn;
end;

procedure TEdgeForm.ToggleEdgeMap(Sender: TObject);
begin

  if gTexture3D.EdgeImg = nil then begin
    Caption := inttostr(CubicEdge(gTexture3D,3))+'ms';
    EdgePanel.Visible := true;
    BiasGainTrackChange(nil);
    //Tex2NII ('c:\edge.nii', gTexture3D);
  end else begin
    SetLengthB(gTexture3D.EdgeImg,0);
    EdgePanel.Visible := false;
    GLForm1.CLUTTimerXOn;
  end;
end;

end.
