unit overlay;

interface
{$H+}
uses
{$IFDEF FPC}LResources, {$ENDIF}
{$IFNDEF Unix} Windows,{$ENDIF}

  OpenGL1x, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Grids, Menus, ExtCtrls,lut, texture_3d_unit,
  ComCtrls, histogram, define_types, nii_reslice,nifti_hdr,scaleimageintensity,CLUT;

type

  { TOverlayForm }

  TOverlayForm = class(TForm)
    Additive1: TMenuItem;
    Additive2: TMenuItem;
    AddOverlay1: TMenuItem;
    BackgroundMaskMenu: TMenuItem;
    Closeoverlays1: TMenuItem;
    InterpolateMenu: TMenuItem;
    LUTdrop: TComboBox;
    ModulateMenu: TMenuItem;
    OverlayColorFromZeroMenu: TMenuItem;
    OverlayMenu1: TMainMenu;
    N0opaque1: TMenuItem;
    N0opaque2: TMenuItem;
    N100transparent1: TMenuItem;
    N100transparent2: TMenuItem;
    N201: TMenuItem;
    N202: TMenuItem;
    N401: TMenuItem;
    N402: TMenuItem;
    N501: TMenuItem;
    N502: TMenuItem;
    N601: TMenuItem;
    N602: TMenuItem;
    N801: TMenuItem;
    N802: TMenuItem;
    Onbackground1: TMenuItem;
    Onotheroverlays1: TMenuItem;
    OptionsMenu1: TMenuItem;
    Panel1: TPanel;
    File1: TMenuItem;
    Transparency1: TMenuItem;
    StatusBar1: TStatusBar;
    StringGrid1: TStringGrid;
    //ModulateMenu: TMenuItem;
    //OverlayColorFromZeroMenu: TMenuItem;

    procedure SetSubmenuWithTag (var lRootMenu: TMenuItem; lTag: Integer);
    procedure BlendOverlays (var lTexture: TTexture);
    procedure UpdateImageIntensity (lOverlay: integer);
    procedure UpdateImageIntensityMinMax (lOverlay: integer; lMinIn,lMaxIn: single);
    procedure FormCreate(Sender: TObject);
    procedure LUTdropChange(Sender: TObject);
    procedure StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    function AddOverlay(lFilename: string; lVolume: integer): integer;
    procedure Addoverlay1Click(Sender: TObject);
    procedure Closeoverlays1Click(Sender: TObject);
    procedure UpdateColorSchema;
    procedure SetBackgroundAlpha(Sender: TObject);
    procedure SetOverlayAlpha(Sender: TObject);
    procedure StatusPanelUpdate;
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure UpdateLUT(lOverlay,lLUTIndex: integer; lChangeDrop: boolean);
    procedure ChangeOverlayUpdate;
    procedure StringGrid1KeyPress(Sender: TObject; var Key: Char);
    procedure ReadCell (ACol,ARow: integer);
    procedure StringGrid1Exit(Sender: TObject);
    procedure InterpolateMenuClick(Sender: TObject);
    procedure BackgroundMaskMenuClick(Sender: TObject);
    procedure StringGrid1DrawCell(Sender: TObject; ACol, ARow: Integer;
              Rect: TRect; State: TGridDrawState);
    procedure CMDialogKey(var msg: TCMDialogKey);
    procedure OverlayColorFromZeroMenuClick(Sender: TObject);
    { Private declarations }
  public
    { Public declarations }
  end;

var
  OverlayForm: TOverlayForm;
const
  kMinOverlayIndex = 1;
  kMaxOverlays = 8;
var
  gEnterCell: boolean = false;
  gOpenOverlays : integer = 0;
  gOverlayImg : array [kMinOverlayIndex..kMaxOverlays] of TMRIcroHdr;
  gOverlayCLUTrec : array [kMinOverlayIndex..kMaxOverlays] of TCLUTrec;
  gPrevCol: integer = 0;
  gPrevRow: integer = 0;
implementation

uses mainunit, readint;
{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}
{$IFDEF UNIX}
const
 vk_Tab = 9;
{$ENDIF}
procedure TOverlayForm.ChangeOverlayUpdate;
begin
  GLForm1.M_Refresh := true;
  GLForm1.UpdateTimer.enabled := true;
end;

procedure GenerateSlice (l32bitOutput: RGBQuadp; l8BitInput: bytep; lLUT: TLUT; lSlicePixels: integer);
var
  lI: integer;
begin
  if lSlicePixels < 1 then
    exit;
  for lI := 1 to lSlicePixels do begin
    l32bitOutput^[lI] := lLUT[l8BitInput^[lI]];
  end;//each voxel
end;

procedure MinMax (var lMin,lMax: integer; lVal: integer);
begin
    if lVal < lMin then
      lMin := lVal;
    if lVal > lMax then
      lMax := lVal;
end;

procedure ModulateBlend (lSlicePixels,lPosIn: integer; var	lBGp,lOverlayP: ByteP);
var
  lMin,lMax,lPixel,lPos: integer;
  lSlope,lWt: single;
begin
    lPos := lPosIn+3;
    lMin := lBGp^[lPos];
    lMax := lBGp^[lPos];
	  for lPixel := 1 to lSlicePixels do begin
		    MinMax(lMin,lMax,lBGp^[lPos]);
        inc(lPos,4);
    end;
    if lMin >= lMax then
      exit;//no range
    lSlope := 1/(lMax-lMin);
    lPos := lPosIn;
	  for lPixel := 1 to lSlicePixels do begin
        lWt := (lBGp^[lPos+3]-lMin)*lSlope;
        //red...
		    lBGp^[lPos] := round(lWt*lOverlayP^[lPos]);
		    inc(lPos);
		    //green...
        lBGp^[lPos] := round(lWt*lOverlayP^[lPos]);
		    inc(lPos);
        //blue...
		    lBGp^[lPos] := round(lWt*lOverlayP^[lPos]);
		    inc(lPos);
        //skip alpha
		    inc(lPos);
	  end;//for each pixel
end;

procedure AlphaBlend32(lBGQuad,lOverlayQuad : RGBQuadp; lBG0Clr,lOverlay0Clr: DWord; lSlicePixels, lOverlayTransPct: integer; lMaskWithBackground: boolean);  // 630
var
	lOverlayByte,lBGwt,lOverlaywt,lPixel,lPos:integer;
	lBGp,lOverlayP: ByteP;
  lOverlayDWordp : DWordp;
begin
	lBGp := ByteP(lBGQuad);
	lOverlayP := ByteP(lOverlayQuad);
	lOverlayDWordp := DWordp(lOverlayQuad);
	//next: transparency weighting
	lBGwt := round((lOverlayTransPct)/100 * 1024);
	lOverlaywt := round((100-lOverlayTransPct)/100 * 1024);
  lOverlayByte := round((255-lOverlayTransPct)/100 * 255);
	//next redraw each pixel
	lPos := 1;
  if lOverlayTransPct > -1 then begin
	  for lPixel := 1 to lSlicePixels do begin
      if lOverlayDWordp^[lPixel] = lOverlay0Clr then
		    inc(lPos,4)
      else if (lOverlayP^[lPos+3] = 0) or ((lBGp^[lPos+3] = 0 )  and (lMaskWithBackground)) then
		    inc(lPos,4)
      else begin
		    lBGp^[lPos] := (lBGp^[lPos]*lBGwt+lOverlayP^[lPos]*lOverlaywt) shr 10;
		    inc(lPos);
		    lBGp^[lPos] := (lBGp^[lPos]*lBGwt+lOverlayP^[lPos]*lOverlaywt) shr 10;
		    inc(lPos);
		    lBGp^[lPos] := (lBGp^[lPos]*lBGwt+lOverlayP^[lPos]*lOverlaywt) shr 10;
		    inc(lPos);
        if (not lMaskWithBackground) and (lBGp^[lPos]< lOverlayByte) then
          lBGp^[lPos] := lOverlayByte;
		    inc(lPos);
      end;
	  end;//for each pixel
  end else if lOverlayTransPct = -2 then begin
    ModulateBlend (lSlicePixels,lPos,lBGp,lOverlayP);
  end else begin //less than one : additive
	  for lPixel := 1 to lSlicePixels do begin
      if lOverlayDWordp^[lPixel] = lOverlay0Clr then
		    inc(lPos,4)
	    else if (lBGp^[lPos+3] = 0 ) and (lMaskWithBackground) then
		    inc(lPos,4)
	    else begin
		    if lOverlayP^[lPos] > lBGp^[lPos] then lBGp^[lPos] := lOverlayP^[lPos];
		    inc(lPos);
		    if lOverlayP^[lPos] > lBGp^[lPos] then lBGp^[lPos] := lOverlayP^[lPos];
		    inc(lPos);
		    if lOverlayP^[lPos] > lBGp^[lPos] then lBGp^[lPos] := lOverlayP^[lPos];
		    inc(lPos);
        if (not lMaskWithBackground) and (lOverlayP^[lPos] > lBGp^[lPos])  then
          lBGp[lPos] := lOverlayP[lPos];
		    inc(lPos);
	    end;
	  end; //for each pixel
  end; //additive
end;

procedure TOverlayForm.BlendOverlays (var lTexture: TTexture);
var
  lOverlaySliceP,lOverlaySlice2P: RGBQuadp;
  lOffset,lRGBOffset: integer;
  lSlicePixels,lSliceBytes,lSlice,lO,lVox: integer;
begin
  lVox := lTexture.FiltDim[1]*lTexture.FiltDim[2]*lTexture.FiltDim[3];
  if (lVox < 1) or (gPrefs.BackgroundAlpha = 100) or (gOpenOverlays < 1) or (lTexture.DataType <> GL_RGBA) then
    exit;
  for lO := 1 to gOpenOverlays do
      if lVox <> gOverlayImg[lO].ScrnBufferItems then
        exit;//error - sizes do not match
  lSlicePixels :=lTexture.FiltDim[1]*lTexture.FiltDim[2];
  lSliceBytes:= lSlicePixels*sizeof(TGLRGBQuad);
  getmem(lOverlaySliceP,lSliceBytes);
  getmem(lOverlaySlice2P,lSliceBytes);
  lOffset := 1;
  lRGBOffset := 0;
  for lSlice := 1 to lTexture.FiltDim[3] do begin
    GenerateSlice(lOverlaySliceP,@gOverlayImg[1].ScrnBuffer^[lOffset],gOverlayImg[1].LUT,lSlicePixels);
    if gOpenOverlays > 1 then begin
      for lO := 2 to gOpenOverlays do begin
        GenerateSlice(lOverlaySlice2P,@gOverlayImg[lO].ScrnBuffer^[lOffset],gOverlayImg[lO].LUT,lSlicePixels);
        AlphaBlend32(lOverlaySliceP,lOverlaySlice2P, DWOrd(gOverlayImg[1].LUT[0]),DWord(gOverlayImg[lO].LUT[0]), lSlicePixels, gPrefs.OverlayAlpha, gPrefs.MaskOverlayWithBackground );  // 630
      end;
    end;
    AlphaBlend32(RGBquadp(@lTexture.FiltImg^[lRGBOffset]),lOverlaySliceP, 0,DWord(gOverlayImg[1].LUT[0]), lSlicePixels, gPrefs.BackgroundAlpha,gPrefs.MaskOverlayWithBackground);  // 630

    lOffset := lOffset + lSlicePixels;
    lRGBOffset := lRGBOffset + lSliceBytes;
  end;
  freemem(lOverlaySliceP);
  freemem(lOverlaySlice2P);
end; //BlendOverlays

procedure InitOverlay (lOverlayIndex: integer);
begin
  gOverlayImg[lOverlayIndex].ImgBufferItems := 0;
  gOverlayImg[lOverlayIndex].ScrnBufferItems := 0;
  gOverlayImg[lOverlayIndex].ImgBufferUnaligned := nil;
  gOverlayImg[lOverlayIndex].ImgBuffer := nil;
  gOverlayImg[lOverlayIndex].ScrnBuffer := nil;
end;

procedure CloseOverlay (lOverlayIndex: integer);
begin
 OverlayForm.LUTdrop.visible := false;
  OverlayForm.StringGrid1.Selection := TGridRect(Rect(-1, -1, -1, -1));
  if gOverlayImg[lOverlayIndex].ImgBufferUnaligned <> nil then
    freemem(gOverlayImg[lOverlayIndex].ImgBufferUnaligned);
  if gOverlayImg[lOverlayIndex].ScrnBuffer <> nil then
    freemem(gOverlayImg[lOverlayIndex].ScrnBuffer);
  InitOverlay(lOverlayIndex);
end;

procedure CloseOverlays;
var
  I: integer;
begin
    for I := kMinOverlayIndex to kMaxOverlays do
      CloseOverlay(I);
    gOpenOverlays := 0;
end;

procedure InitOverlays;
var
  I: integer;
begin
    gOpenOverlays := 0;
    for I := kMinOverlayIndex to kMaxOverlays do
      InitOverlay(I);
end;

procedure TOverlayForm.UpdateColorSchema;
//var
//	lSearchRec: TSearchRec;
begin
  LUTdrop.Items.Clear;
  LUTdrop.Items.Add('Grayscale');
  LUTdrop.Items.Add('Red');
  LUTdrop.Items.Add('Green');
  LUTdrop.Items.Add('Blue');
  LUTdrop.Items.Add('Violet [r+b]');
  LUTdrop.Items.Add('Yellow [r+g]');
  LUTdrop.Items.Add('Cyan [g+b]');
  UpdateColorSchemes(LUTdrop);
  (*if FindFirst(CLUTdir+pathdelim+'*.clut', faAnyFile, lSearchRec) = 0 then
	 repeat
		  LUTdrop.Items.Add(ParseFileName(ExtractFileName(lSearchRec.Name)));
	 until (FindNext(lSearchRec) <> 0);
  FindClose(lSearchRec);*)
end;//UpdateColorSchemes

procedure TOverlayForm.SetSubmenuWithTag (var lRootMenu: TMenuItem; lTag: Integer);
var
	lCount,lSubMenu: integer;
begin
	lCount := lRootMenu.Count;
	if lCount < 1 then exit;
	for lSubMenu := (lCount-1) downto 0 do
		if lRootmenu.Items[lSubmenu].Tag = lTag then begin
			lRootmenu.Items[lSubmenu].Checked := true;
			exit
		end;
	//will exit unless tag not found: default select 1st item
	lRootmenu.Items[0].Checked := true;
	//While Recent1.Count > 0 do Recent1.Items[0].Free;
end;



procedure TOverlayForm.FormCreate(Sender: TObject);
var i: Integer;
begin
{$IFDEF Darwin}
  AddOverlay1.ShortCut := ShortCut(Word('A'), [ssMeta]);
{$ENDIF}
StringGrid1.Selection := TGridRect(Rect(-1, -1, -1, -1));
//StringGrid1.Options:= StringGrid1.Options - [goRowSelect] ;
  InitOverlays;
  StringGrid1.DefaultRowHeight := LUTdrop.Height;
  LUTdrop.Visible := False;
  UpdateColorSchema;
  StringGrid1.Cells[0, 0] := 'Filename';
  StringGrid1.Cells[1, 0] := 'Colorscheme';
  StringGrid1.Cells[2, 0] := 'Min';
  StringGrid1.Cells[3, 0] := 'Max';
//  StringGrid1.Selection:=TGridRect(Rect(-1,-1,-1,-1));
  for i := 1 to StringGrid1.RowCount-1 do
    StringGrid1.Cells[0, i] := IntToStr(i);
  SetSubmenuWithTag(Onotheroverlays1,gPrefs.OverlayAlpha);
  SetSubmenuWithTag(Onbackground1,gPrefs.BackgroundAlpha);
  //StringGrid1.Selection := Rect(-1,-1,-1,-1);

end;

procedure TOverlayForm.CMDialogKey(var msg: TCMDialogKey);
begin
(*  if (ActiveControl = LUTdrop) then begin
{$IFNDEF FPC}
    if (msg.CharCode = VK_TAB) then begin
      //set focus back to the grid and pass the tab key to it
      //LUTdrop.SetFocus;
      {$IFNDEF UNIX}
      //LUTdrop.Perform(WM_KEYDOWN, msg.CharCode, msg.KeyData);
      {$ENDIF}
      // swallow this message
      msg.result := 1;
      Exit;
    end;
//laz bug
{$ENDIF}
  end;
  *)
  inherited;
end;

procedure TOverlayForm.LUTdropChange(Sender: TObject);
var intRow: Integer;
begin
  inherited;
  if Lutdrop.Tag < 1 then
     exit;
  intRow := StringGrid1.Row;
  if intRow < 0 then
    intRow := Lutdrop.Tag;
  if (intRow < 1) or (intRow > kMaxOverlays) then
    exit;
  StringGrid1.Cells[1, intRow] := LutDrop.Items[LUTdrop.ItemIndex];
  UpdateLUT(intRow,LUTdrop.ItemIndex,false);
  ChangeOverlayUpdate;
  StringGrid1.Selection:=TGridRect(Rect(-1,-1,-1,-1));
end;

function HasDigit (var lS: string): boolean;
//do not attempt to convert '-', '.', or '-.' as a number...
var
   lI,lLen: integer;
begin
     result := false;
     lLen := length (lS);
     if lLen < 1 then
        exit;
     for lI := 1 to lLen do begin
         if lS[lI] in ['0'..'9'] then begin
            result := true;
            exit;
         end;
     end;
end;

procedure TOverlayForm.ReadCell (ACol,ARow: integer);
var
  lF: single;
  lS: string;
begin
  if (ARow < StringGrid1.FixedRows) or (ARow > kMaxOverlays) then
    exit;
  if (ACol <> 2) and (ACol <> 3) then
    exit;
  lS := StringGrid1.Cells[ACol,ARow];
  if not HasDigit(lS) then
    exit;
  try
       lF := strtofloat(lS);
  except
          exit;
          {on EConvertError do begin
             Msg('Unable to convert the string '+lStr+' to a number');
             result := 1;
             exit;
          end;}
  end; {except}
  if ACol = 2 then
    gOverlayImg[ARow].WindowScaledMin := lF
  else
    gOverlayImg[ARow].WindowScaledMax := lF;

  UpdateImageIntensity(ARow);
end;

procedure TOverlayForm.StringGrid1SelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
var R: TRect;
begin
   if not OverlayForm.visible then
      exit;
  if (ACol < 1) or (ACol > 3) then
    exit;
  if (ARow < 1) or (ARow > gOpenOverlays) then
     exit;
  ReadCell(gPrevCol,gPrevRow);
  if (ACol = 1) and  (ARow <> 0) then begin
    {Size and position the combo box to fit the cell}
    R := StringGrid1.CellRect(ACol, ARow);
    R.Left := R.Left + StringGrid1.Left;
    R.Right := R.Right + StringGrid1.Left;
    R.Top := R.Top + StringGrid1.Top;
    R.Bottom := R.Bottom + StringGrid1.Top;
    {Show the combobox}
    with LUTdrop do begin
      Tag := 0;
      Left := R.Left + 1;
      Top := R.Top + 1;
      Width := (R.Right + 1) - R.Left;
      Height := (R.Bottom + 1) - R.Top;
      ItemIndex := Items.IndexOf(StringGrid1.Cells[ACol, ARow]);
      Visible := True;
      SetFocus;
      Tag := ARow;
    end;
  end else begin
      LUTdrop.visible := false;
      ReadCell(ACol,ARow);
      gEnterCell := true;
  end;
  CanSelect := True;
end;

procedure TOverlayForm.UpdateLUT(lOverlay,lLUTIndex: integer; lChangeDrop: boolean);
begin
  if gOpenOverlays > kMaxOverlays then
    exit;
  if lLUTIndex >= LUTdrop.Items.Count then
    gOverlayImg[lOverlay].LUTindex:= 0
  else
    gOverlayImg[lOverlay].LUTindex:= lLUTIndex;
  if lChangeDrop then begin
    StringGrid1.Cells[1, lOverlay] := LUTdrop.Items[gOverlayImg[lOverlay].LUTindex];
    //LUTdrop.ItemIndex := gOverlayImg[lOverlay].LUTindex;
  end;
  LUTdropLoad(gOverlayImg[lOverlay].LUTindex, gOverlayImg[lOverlay].LUT, LUTdrop.Items[lLUTindex], gOverlayCLUTrec[lOverlay]);
end;

procedure TOverlayForm.UpdateImageIntensity (lOverlay: integer);
begin
  RescaleImgIntensity(gOverlayImg[lOverlay] );
  //StatusBar1.Panels[0].text := 'x '+inttostr(lOverlay)+' '+floattostr(gOverlayImg[lOverlay].WindowScaledMin)+'...'+floattostr(gOverlayImg[lOverlay].WindowScaledMax) ;
  ChangeOverlayUpdate;
end;

procedure TOverlayForm.UpdateImageIntensityMinMax (lOverlay: integer; lMinIn,lMaxIn: single);
var
   lMin,lMax: single;
begin
     if lMinIn > lMaxin then begin
        lMin := lMaxIn;
        lMax := lMinIn;
     end else begin
         lMin := lMinIn;
         lMax := lMaxIn;
     end;
    gOverlayImg[lOverlay].WindowScaledMin := lMin;
    gOverlayImg[lOverlay].WindowScaledMax := lMax;
    StringGrid1.Cells[2,lOverlay] := floattostr(lMin);
    StringGrid1.Cells[3,lOverlay] := floattostr(lMax);
        UpdateImageIntensity(lOverlay);
end;

function TOverlayForm.Addoverlay(lFilename: string; lVolume: integer): integer;
var
  lL: integer;
  lV,lFilenameX: string;
begin
  result := -1;
  StringGrid1.Selection := TGridRect(Rect(2, 3, 2, 3));
  if gOpenOverlays >= kMaxOverlays then begin
      showmessage('Too many overlays open. Please close an overlay before adding a new one.');
      exit;
  end;
  if (gTexture3D.FiltDim[1] < 1) or (gTexture3D.FiltDim[2] < 1) or (gTexture3D.FiltDim[3] < 1) then begin
    showmessage('Please load a background image before loading an overlay.');
    exit;
  end;
  lFilenameX := lFilename;
  GLForm1.CheckFilename (lFilenameX,false);
  if not fileexists(lFilenameX) then begin
    showmessage('Unable to find overlay named '+lFilename);
    exit;
  end;
  StringGrid1.Visible := true;
  inc (gOpenOverlays);
    StringGrid1.RowCount := StringGrid1.FixedRows+gOpenOverlays;
  if Reslice2Targ (lFilenameX, gTexture3D.NIFTIhdr,gOverlayImg[gOpenOverlays],gPrefs.InterpolateOverlays,lVolume) = '' then begin
    showmessage('Error loading overlay.');
    dec (gOpenOverlays);
    exit;
  end;
  ComputeThreshM (gOverlayImg[gOpenOverlays]);
  if (gOverlayImg[gOpenOverlays].RGB)  then begin //RGB images
    gOverlayImg[gOpenOverlays].AutoBalMinUnscaled := 0.1;
    gOverlayImg[gOpenOverlays].AutoBalMaxUnscaled := 255;
    if (lVolume mod 3) = 1 then //green
      lL := 2
    else if (lVolume mod 3) = 2 then  //blue
      lL := 3
     else //Red
      lL := 1;
    UpdateLUT(gOpenOverlays,lL,true);
  end else //not RGB
    UpdateLUT(gOpenOverlays,gOpenOverlays,true);
//  StringGrid1.RowCount := StringGrid1.FixedRows+gOpenOverlays;
  if lVolume > 1 then
    lV := ':'+inttostr(lVolume)
  else
    lV := '';
  StringGrid1.Cells[0, gOpenOverlays] := parsefilename(extractfilename(lFilenameX))+lV;
  //if gOverlayImg[gOpenOverlays].LUTindex < LUTdrop.Items.Count then
  //  StringGrid1.Cells[1, gOpenOverlays] := LUTdrop.Items[gOverlayImg[gOpenOverlays].LUTindex];
  //xxComputeMinMax(var lHdr: TTexture);


  StringGrid1.Cells[2, gOpenOverlays] := floattostr(gOverlayImg[gOpenOverlays].AutoBalMinUnscaled);
  StringGrid1.Cells[3, gOpenOverlays] := floattostr(gOverlayImg[gOpenOverlays].AutoBalMaxUnscaled);
  gOverlayImg[gOpenOverlays].WindowScaledMin := gOverlayImg[gOpenOverlays].AutoBalMinUnscaled;
  gOverlayImg[gOpenOverlays].WindowScaledMax := gOverlayImg[gOpenOverlays].AutoBalMaxUnscaled;
  gOverlayImg[gOpenOverlays].LutFromZero := gPrefs.OverlayColorFromZero;
  UpdateImageIntensity (gOpenOverlays);
  ChangeOverlayUpdate;
    LUTdrop.Visible := false;
  result :=gOpenOverlays;
  //RGB
  if (gOverlayImg[gOpenOverlays].RGB) then begin
    if (lVolume < 3) then
      result :=Addoverlay(lFilename, lVolume+1)
    else
     Additive2.Click;//additive
  end; //RGB
end;

procedure TOverlayForm.Addoverlay1Click(Sender: TObject);
var
  lFilename: string;
  lnVol: integer;
begin
  if gOpenOverlays >= kMaxOverlays then begin
      showmessage('Too many overlays open. Please close an overlay before adding a new one.');
      exit;
  end;
      StringGrid1.Selection := TGridRect(Rect(-1, -1, -1, -1));
  if (gTexture3D.FiltDim[1] < 1) or (gTexture3D.FiltDim[2] < 1) or (gTexture3D.FiltDim[3] < 1) then begin
    showmessage('Please load a background image before loading an overlay.');
    exit;
  end;
  if not GLForm1.OpenDialog1.Execute then exit;
  lFilename :=  GLForm1.OpenDialog1.Filename;
  lnVol :=  NIFTIvolumes(lFilename);
  if lnVol < 1 then
      exit;
  if lnVol > 1 then
      lnVol := ReadIntForm.GetInt('4D image: select volume',1,1,lnVol);
  AddOverlay(lFilename,lnVol);
end;

procedure TOverlayForm.Closeoverlays1Click(Sender: TObject);
begin
  StringGrid1.Visible := false;
  LUTdrop.Visible := false;
  CloseOverlays;
  gOpenOverlays := 0;
  StringGrid1.RowCount := StringGrid1.FixedRows+1;
  ChangeOverlayUpdate;
end;

procedure TOverlayForm.StatusPanelUpdate;
begin
    StatusBar1.Panels[0].text := 'Background A '+inttostr(gPrefs.BackgroundAlpha)+'  Overlay A '+inttostr(gPrefs.OverlayAlpha) ;
end;

procedure TOverlayForm.SetBackgroundAlpha(Sender: TObject);
begin
  gPrefs.BackgroundAlpha := (Sender as TMenuItem).tag;
  StatusPanelUpdate;
  ChangeOverlayUpdate;
end;

procedure TOverlayForm.SetOverlayAlpha(Sender: TObject);
begin
  gPrefs.OverlayAlpha := (Sender as TMenuItem).tag;
  StatusPanelUpdate;
  ChangeOverlayUpdate;
end;

procedure TOverlayForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  CloseOverlays;
end;

function IsDigit (letter : char) : boolean;
//  If letter is a digit, 0 through 9, true is returned.
//  Otherwise, false is returned.
begin
  if ((letter <= '9') and (letter >= '0')) then
    IsDigit := true
  else
    IsDigit := false;
end;

procedure TOverlayForm.StringGrid1KeyPress(Sender: TObject; var Key: Char);
const
  EnterKey = #13;
  BackspaceKey = #8;
  ControlC = #3;   //  Copy
  ControlV = #22;  //  Paste
var
  ACol,ARow: integer;
  S: string;
begin
  ACol := abs(StringGrid1.Selection.Right);
  ARow := abs(StringGrid1.Selection.Top);
  gPrevCol := ACol;
  gPrevRow := ARow;
  if (not (IsDigit (Key) or (Key = decimalseparator) or (Key = '+') or (Key = '-') or
        (Key = ControlC) or (Key = ControlV) or (Key = BackspaceKey) or
        (Key = EnterKey))) then begin
    Key := #0;
    exit;
  end;
  if (Key = kTab) then
    exit;
  if (Key = kTab) or (Key = kCR) then begin
    ReadCell(gPrevCol,gPrevRow);
    exit;
  end;

	if(( StringGrid1.Selection.Top = StringGrid1.Selection.Bottom ) and
		( StringGrid1.Selection.Left = StringGrid1.Selection.Right )) then begin
        if gEnterCell then begin
           S := ''
        end else
			  S := StringGrid1.Cells[ StringGrid1.Selection.Left,StringGrid1.Selection.Top ] ;
        gEnterCell := false;
        if ( ( Key = kDEL ) or ( Key = kBS ) )then begin
            if( length( S ) > 0 ) then begin
                setlength( S, length( S ) - 1 ) ;
            end;
        end else
			    S := S + Key ;
		//StringGrid1.Cells[ StringGrid1.Selection.Left, StringGrid1.Selection.Top ] := S ;
   {$IFDEF FPC} StringGrid1.Cells[ StringGrid1.Selection.Left,StringGrid1.Selection.Top ] := S;
    {$ENDIF}
	end ;
end;

procedure TOverlayForm.StringGrid1Exit(Sender: TObject);
begin
  ReadCell(gPrevCol,gPrevRow);
end;

procedure TOverlayForm.InterpolateMenuClick(Sender: TObject);
begin
  gPrefs.InterpolateOverlays := InterpolateMenu.checked;
end;

procedure TOverlayForm.FormShow(Sender: TObject);
begin
  InterpolateMenu.checked := gPrefs.InterpolateOverlays;
  OverlayColorFromZeroMenu.checked := gPrefs.OverlayColorFromZero;
end;

procedure TOverlayForm.OverlayColorFromZeroMenuClick(Sender: TObject);
var
  lO: integer;
begin
  gPrefs.OverlayColorFromZero := OverlayColorFromZeroMenu.checked;
  if  gOpenOverlays < 1 then
    exit;
  for lO := 1 to gOpenOverlays do
    gOverlayImg[lO].LutFromZero := gPrefs.OverlayColorFromZero;
  for lO := 1 to gOpenOverlays do
    RescaleImgIntensity(gOverlayImg[lO] );
  ChangeOverlayUpdate;
end;

procedure TOverlayForm.BackgroundMaskMenuClick(Sender: TObject);
begin
  gPrefs.MaskOverlayWithBackground := BackgroundMaskMenu.checked;
  ChangeOverlayUpdate;
end;


procedure TOverlayForm.StringGrid1DrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
{$IFDEF FPC}
begin
  //This code only works with Delphi
end;
{$ELSE}
var
  dx: Integer;
begin
  with (Sender as TStringGrid) do
  begin
    // Don't change color for first Column, first row
    if (ACol = 0) or (ARow = 0) then
      Canvas.Brush.Color := clBtnFace
    else
    begin
      case ACol of
        1: Canvas.Font.Color := clBlack;
        2: Canvas.Font.Color := clBlue;
      end;
      // Draw the Band
      if ARow mod 2 = 1 then
        Canvas.Brush.Color := $00E1FFF9
      else
        Canvas.Brush.Color := $00FFEBDF;
      Canvas.TextRect(Rect, Rect.Left + 2, Rect.Top + 2, cells[acol, arow]);
      Canvas.FrameRect(Rect);
    end;
  end;
end;
{$ENDIF}


initialization
{$IFDEF FPC}
  {$I overlay.lrs}
{$ENDIF}
end.
