unit slices2D;
 {$include options.inc}
interface
{$H+}
uses
{$IFDEF Unix} {$ELSE} Windows,{$ENDIF}
{$IFDEF GLSCENEOLD}
GLMisc,
{$ELSE}
GLRenderContextInfo,
{$ENDIF}
Classes,OpenGL1x, GLObjects,texture_3d_unit, Controls,
 GLTexture, GLContext,
 GLCanvas,  ComCtrls, ToolWin, define_types,coordinates, vectortypes,GraphicsMathLibrary
  //GLHUDObjects, StdCtrls, Buttons,
  //GLBitmapFont, GLWindowsFont
;

const
  kMaxMosaicDim = 12; //e.g. if 12 then only able to draw up to 12x12 mosaics [=144 slices]
  kEmptyOrient = 0;
  kAxialOrient = 1;
  kCoronalOrient = 2;
  kSagRightOrient = 3;
  kSagLeftOrient = 4;

//procedure Mosaic (var lMosaicString: string);
function SliceMM (lSliceFrac: single; lOrient: integer): single;
procedure MosaicGL (var lMosaicString: string; var lTexture3D: TTexture; var lHtx, lWidx: integer; NoDraw: boolean; var rci: TRenderContextInfo);
//procedure Mosaic (var lMosaicString: string; var lTexture3D: TTexture);
procedure go2D;
procedure end2D;

implementation
uses mainunit, sysutils, dialogs {test only},mosaicprefs;

procedure go2D;
begin
    glColor4f (1, 1, 1,1);
    // Disable depth testing
    glDisable(GL_DEPTH_TEST);
    //Select The Projection Matrix
    glMatrixMode(GL_PROJECTION);
    // Store The Projection Matrix
    glPushMatrix();
    //Reset The Projection Matrix
    glLoadIdentity();
    // Set Up An Ortho Screen
    gluOrtho2D(0.0, 1.0, 0, 1.0);
    //Select The Modelview Matrix
    glMatrixMode(GL_MODELVIEW);
    // Stor the Modelview Matrix
    glPushMatrix;
    // Reset The Modelview Matrix
    glLoadIdentity;
end;

procedure end2D;
begin
    glColor4f (1, 1, 1,1);
    //Select The Projection Matrix
    glMatrixMode(GL_PROJECTION);
    // Restore The Old Projection Matrix
    glPopMatrix;
    //Select the Modelview Matrix
    glMatrixMode(GL_MODELVIEW);
    //Restore the Old Projection Matrix
    glPopMatrix;
    //Re-enable Depth Testing
    glEnable(GL_DEPTH_TEST);
end;

type
  TMosaicSlices = array [1..kMaxMosaicDim, 1..kMaxMosaicDim] of single;
  TMosaicOrient = array [1..kMaxMosaicDim, 1..kMaxMosaicDim] of byte;
  TMosaicText = array [1..kMaxMosaicDim, 1..kMaxMosaicDim] of boolean;
  TMosaic =record
    Slices: TMosaicSlices;
    Orient: TMosaicOrient;
    Text: TMosaicText;
    HOverlap,VOverlap: single;
  end;

function SliceToFrac (lSlice: single; lSlices : integer): single;
//indexed from 1
begin
    result := 0.5;
    if lSlices < 1 then
      exit;
    //result := round(((lFrac+ (1/lSlices))*lSlices));

    result := (lSlice-1)/lSlices;
    //bound in range 0..1
    if result < 0 then
      result := 0;
    if result > 1 then
      result := 1;
end;

function SliceFracF (lSliceMM: single; lOrient: integer; var lInvMat: TMatrix): single;
var
  X,Y,Z: single;

begin
    X := 0;
    Y := 0;
    Z := 0;
    case lOrient of
      kAxialOrient : Z := lSLiceMM;
      kCoronalOrient : Y := lSliceMM;
      kSagRightOrient : X := lSliceMM;
      kSagLeftOrient : X := lSliceMM;
    end;
    mm2Voxel (X,Y,Z, lInvMat);
    //-1 as the values are indexed from zero...
    //X := SliceToFrac(X-1,gTexture3D.FiltDim[1]);
    //Y := SliceToFrac(Y-1,gTexture3D.FiltDim[2]);
    //Z := SliceToFrac(Z-1,gTexture3D.FiltDim[3]);
    case lOrient of
      kAxialOrient : result := SliceToFrac(Z,gTexture3D.FiltDim[3]);
      kCoronalOrient : result := SliceToFrac(Y,gTexture3D.FiltDim[2]);
      kSagRightOrient,kSagLeftOrient : result := SliceToFrac(X,gTexture3D.FiltDim[1]);
      else result := 0; //should be impossible - prevents compiler warning
    end;
end; //SliceMM


procedure StereoTaxicSpaceToFrac (var lRows, lCols: integer; var lMosaic: TMosaic);
var
  lRow,lCol: integer;
  lFrac: boolean;
  lInvMat: TMatrix;
  lOK: boolean;
begin
  if (lCols < 1) or (lRows < 1)  then
    exit;
  //fx(lMosaic.Slices[1,1]);
  lFrac := true;
  for lRow := 1 to lRows do
    for lCol := 1 to lCols do
      if (lMosaic.Orient[lCol,lRow] <> kEmptyOrient) and ((lMosaic.Slices[lCol,lRows] < 0) or (lMosaic.Slices[lCol,lRows] > 1)) then
        lFrac := false;
  if lFrac then exit;
  lInvMat := Hdr2InvMat (gTexture3D.NIftiHdr,lOK);
  if not lOK then exit;
  for lRow := 1 to lRows do begin
    for lCol := 1 to lCols do begin
      if lMosaic.Orient[lCol,lRow] <> kEmptyOrient then
        lMosaic.Slices[lCol,lRow] := SliceFracF (lMosaic.Slices[lCol,lRow], lMosaic.Orient[lCol,lRow],lInvMat)
    end;//col
  end;//row
end;

procedure ReadMosaicStr (var lMosaicString: string; var lRows, lMaxCol: integer; var lMosaic: TMosaic);
// '0.2 0.4 0.8; 0.9' has 3 columns and 2 rows
// '0.2 0.4; 0.8 0.9' has 2 columns and 2 rows
// '0.2 0.4; 0.8 s 0.5' has 2 columns and 2 rows, with final item a sagittal slice
//INPUTS
// a c s z   --- orientations (axial,coronal, sagittal,sagittal[mirror]
//  numbers, including decimal separator and E (for scientific notation)
//  ;        ---- next row
// v h     ---- vertical and horizontal overlap, e.g. 0 mean snone, while 0.2 is 20% range is -1..1
//  note v and h must be followed by a number
function S2F (lStr: String): single;//float to string
begin
  try
    result := StrToFloat(lStr);    // Hexadecimal values are not supported
  except
    //report problem here..
    result := 1;
  end;//try ... except
end; //nested S2F
procedure DummyMosaic;
begin
    lRows := 2;
    lMaxCol := 2;
    lMosaic.Slices[1,1] := 0.2;
    lMosaic.Slices[1,2] := 0.4;
    lMosaic.Slices[2,1] := 0.6;
    lMosaic.Slices[2,2] := 0.8;
end; //nested proc DummyMosaic
var
  lDone: boolean;
  lCh,lCh2: char;
  lNumStr: string;
  lFloat: single;
  lX,lY,lPos,lLen,lCurrentOrient,lCol : integer;
  lCurrentText : boolean;
begin
  lCurrentOrient := kAxialOrient;
  lCurrentText := false;
  for lX := 1 to kMaxMosaicDim do begin
    for lY := 1 to kMaxMosaicDim do begin
      lMosaic.Orient[lX,lY] := kEmptyOrient;
      lMosaic.Text[lX,lY] := lCurrentText;
    end;
  end;
  lMosaic.HOverlap := 0;
  lMosaic.VOverlap := 0;
  lRows := 1;
  lMaxCol := 0;
  lLen := length(lMosaicString);
  if lLen < 1 then begin
    DummyMosaic;
    exit;
  end;
  lPos := 1;
  lNumStr := '';
  //lFloat := -1;
  lCol := 0;
  while lPos <= lLen do begin
    lCh := upcase(lMosaicString[lPos]);
    if lCh in ['0'..'9','-', '+','E',DecimalSeparator] then
      lNumStr := lNumStr + lCh;
    if (lPos = lLen) or (not (lCh in ['0'..'9','-', '+','E',DecimalSeparator])) then begin //not number
      //first, read previous nuber
      if lNumStr <> '' then begin
        lCol := lCol + 1;
        lMosaic.Slices[lCol,lRows] := S2F(lNumStr);
        //if lMosaic.Slices[lCol,lRows] < 1 then
        //  fx(lMosaic.Slices[lCol,lRows]);
        //showmessage(floattostr(lMosaic.Slices[lCol,lRows])+'  '+lNumStr);
        lMosaic.Orient[lCol,lRows] := lCurrentOrient;
        lMosaic.Text[lCol,lRows] := lCurrentText;
        //if lCurrentText then GLForm1.caption := 'xxx'+'  '+inttostr(random(222));
      end;//if lNumStr <> ''
      lNumStr := '';
      //next - see if this is some other command, else whitespace
      if lCh = 'A' then
        lCurrentOrient := kAxialOrient;
      if lCh = 'C' then
        lCurrentOrient := kCoronalOrient;
      if lCh = 'S' then
        lCurrentOrient := kSagRightOrient;
      if lCh = 'Z' then
        lCurrentOrient := kSagLeftOrient;
      if lCh = 'L' then begin
        lCurrentText := True;
        //GLForm1.caption := 'xxx'+lMosaicString[lPos+1] +'  '+inttostr(random(222));
        if (lPos < lLen) and (lMosaicString[lPos+1] = '-') then begin
          lCurrentText := False;
          inc(lPos);
        end else if (lPos < lLen) and (lMosaicString[lPos+1] = '+') then
          inc(lPos);


      end;
      if (lCh = 'V') or (lCh = 'H') then begin
        lDone := false;
        repeat
          lCh2 := upcase(lMosaicString[lPos]);
          if lCh2 in ['0'..'9','-', '+','E',DecimalSeparator] then
            lNumStr := lNumStr + lCh2
          else if lNumStr <> '' then begin
            lDone := true;
            dec(lPos);
          end;
          inc(lPos)
        until (lPos > lLen) or lDone;
        lFloat := S2F(lNumStr);
        if (lFloat > 1) or (lFloat < -1) then
          lFloat := 0;
        if lCh = 'V' then
          lMosaic.VOverlap := lFloat
        else
          lMosaic.HOverlap := lFloat;
        lNumStr := '';
      end; //V or H ... read overlap
      if lCh = ';' then begin
        lRows := lRows + 1;
        if lCol > lMaxCol then begin
          lMaxCol := lCol;
        end;
        lCol := 0;
      end;
    end; //not number
    inc(lPos);//read next character
  end; //for each character in string
  //net - last row may not have a ; so check if final line has most columns...
  if lCol > lMaxCol then
    lMaxCol := lCol;
  StereoTaxicSpaceToFrac (lRows, lMaxCol, lMosaic)
end; //proc ReadMosaicStr

function FracToSlice (lFrac: single; lSlices : integer): single;
//indexed from 1
begin
    //result := trunc(lFrac*lSlices)+1;

    //result := round((lFrac*lSlices)+ (1/lSlices));
    result := round(((lFrac+ (1/lSlices))*lSlices));
    //result := round(lFrac*lSlices);
    if result > lSLices then //e.g. if lFrac = 1.0, then result is lSlices+1!
      result := lSlices;
    if result < 1 then //This is impossible if lFrac is 0..1, but just in case...
      result := 1;
end;

function SliceMM (lSliceFrac: single; lOrient: integer): single;
var
  X,Y,Z: single;
begin
    X := 0.5;
    Y := 0.5;
    Z := 0.5;
    case lOrient of
      kAxialOrient : Z := lSLiceFrac;
      kCoronalOrient : Y := lSliceFrac;
      kSagRightOrient : X := lSliceFrac;
      kSagLeftOrient : X := lSliceFrac;
    end;
    X := FracToSlice(X,gTexture3D.FiltDim[1]);
    Y := FracToSlice(Y,gTexture3D.FiltDim[2]);
    Z := FracToSlice(Z,gTexture3D.FiltDim[3]);
    //GLForm1.Caption := inttostr(round(X))+' '+inttostr(round(Y))+' '+inttostr(round(Z)) ;
    Voxel2mm(X,Y,Z,gTexture3D.NIfTIHdr);
    case lOrient of
      kAxialOrient : result := Z;
      kCoronalOrient : result := Y;
      kSagRightOrient,kSagLeftOrient : result := X;
      else result := 0; //should be impossible - prevents compiler warning
    end;
end; //SliceMM

procedure DrawText(var glc: TGLCanvas; lSlice: single; lL,lT,lOrient,lDec: integer; var rci: TRenderContextInfo);
var
  lSmm: single;
  lS: string;
  lW,lH: integer;
begin

    lSmm := SliceMM (lSlice, lOrient);
    lS := realtostr(lSmm,lDec);
    //lS := realtostr(lSlice,1);
    lW := (GLForm1.GLWindowsBitmapFont1.CalcStringWidth(lS)+1) div 2 + 1;
    lH := GLForm1.GLWindowsBitmapFont1.Font.Size * 2;
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable (GL_BLEND);
      glBegin(GL_POLYGON);
        glColor4ub(gPrefs.TextBorder.rgbRed,gPrefs.TextBorder.rgbGreen,gPrefs.TextBorder.rgbBlue,gPrefs.TextBorder.rgbReserved);
        glVertex2f(lL-lW,lT);
        glVertex2f(lL-lW,lT+lH);
        glVertex2f(lL,lT+lH+(lH div 2));
        glVertex2f(lL+lW,lT+lH);
        glVertex2f(lL+lW,lT);
      glEnd;//POLYGON
    GLForm1.GLWindowsBitmapFont1.TextOut(rci, lL-lW+1, lT, lS,RGBA2TColor(gPrefs.TextColor) { $FFFFFF});
end;

procedure DrawSlice(lSlice: single; lL,lT,lScrnW,lScrnH: integer; lMosaicOrient: byte; var lTexture3D: TTexture);
var
  lH,lW,lZoom,lTs,lLs,lI: single;
begin
  lZoom := 1;
  lLs := lL/ lScrnW;
  lTs := 1-(lT/ lScrnH);
  //showmessage(floattostr(lSlice)+'  '+inttostr(lL)+'  '+inttostr(lT));
  if lMosaicOrient = kCoronalOrient then begin
    lH := lTexture3D.dim[3];//coronal image - height is Z axis
    lW := lTexture3D.dim[1];//cornal image - width is X axis
    lH := lH/lScrnH*lZoom;
    lW := lW/lScrnW*lZoom;
    glBegin(GL_QUADS);
      glTexCoord3d (0, lSlice, 1);
      glVertex2f(lLs,lTs);
      glTexCoord3d ( 0,lSlice, 0);
      glVertex2f(lLs,lTs-lH);
      glTexCoord3d (1,lSlice,0);
      glVertex2f(lLs+lW,lTs-lH);
      glTexCoord3d (1,lSlice, 1);
      glVertex2f(lLs+lW,lTs);
    glend;
  end else if lMosaicOrient = kAxialOrient then begin
    lH := lTexture3D.dim[2];//axial image - height is Y axis
    lW := lTexture3D.dim[1];//axial image - width is X axis
    lH := lH/lScrnH*lZoom;
    lW := lW/lScrnW*lZoom;
    glBegin(GL_QUADS);
      glTexCoord3d(1, 1, lSlice);
      glVertex2f(lLs+lW,lTs);
      glTexCoord3d (0, 1, lSlice);
      glVertex2f(lLs,lTs);
      glTexCoord3d(0,0, lSlice);
      glVertex2f(lLs,lTs-lH);
      glTexCoord3d(1, 0, lSlice);
      glVertex2f(lLs+lW,lTs-lH);
    glEnd;
  end else begin //Sagittal
    lH := lTexture3D.dim[3];//coronal image - height is Z axis
    lW := lTexture3D.dim[2];//cornal image - width is Y axis
    lH := lH/lScrnH*lZoom;
    lW := lW/lScrnW*lZoom;
    if lMosaicOrient = kSagRightOrient then //choose left or right profile...
      lI := 1
    else
      lI := 0;
    glBegin(GL_QUADS);
      glTexCoord3d (lSlice, 1-lI, 1);
      glVertex2f(lLs,lTs);
      glTexCoord3d (lSlice, 1-lI, 0);
      glVertex2f(lLs,lTs-lH);
      glTexCoord3d (lSlice,lI,0);
      glVertex2f(lLs+lW,lTs-lH);
      glTexCoord3d (lSlice, lI, 1);
      glVertex2f(lLs+lW,lTs);
    glend;
  end;//if.. else Sag
end; //Proc DrawSlice

procedure MosaicGL (var lMosaicString: string; var lTexture3D: TTexture; var lHtx, lWidx: integer; NoDraw: boolean; var rci: TRenderContextInfo);
var
  lDec,lOffset,lRows, lCols, lX, lY, lMaxWid,lMaxHt,lWid,lWx,lT,lH: integer;
  lMosaic: TMosaic;
  lSliceMM,lMaxMM,lMinMM: single;
  lScrnW,lScrnH: integer;
  glc : TGLCanvas;
begin
  lMaxMM := 0;
  ReadMosaicStr (lMosaicString,lRows, lCols, lMosaic);
  lMaxHt :=0;
  if (lCols < 1) or (lRows < 1)  then
    exit;
  //next prepare screen
  lScrnH := GLForm1.GLSceneViewer1.height;
  if lScrnH = 0 then
    lScrnH := 1;
  lScrnW := GLForm1.GLSceneViewer1.width;
  if lScrnW = 0 then
    lScrnW := 1;
(*for lY := 1 to lRows do
  for lX := 1 to lCols do
      if lMosaic.Text[lX,lY] then
      GLForm1.caption := 'xxx' +'  '+inttostr(random(222));
  *)
  //next - determine width and height of image (in pixels)
  //first pass - compute dimensions, second pass draw images
  lMaxWid := 0;
  lT := 0;


  for lY := 1 to lRows do begin
    lMaxHt := 0;
    lWid := 0;
    for lX := 1 to lCols do begin
      //determine width
      if (lMosaic.Orient[lX,lY] = kEmptyOrient) then
        lWx := 0
      else if (lMosaic.Orient[lX,lY] = kAxialOrient) or (lMosaic.Orient[lX,lY] =kCoronalOrient) then
        lWx := lTexture3D.Dim[1] //width left-right
      else
        lWx := lTexture3D.Dim[2]; //sagittal width anterior-posterion;
      //compute height
      if (lMosaic.Orient[lX,lY] = kEmptyOrient) then
        lH := 0
      else if (lMosaic.Orient[lX,lY] = kAxialOrient)  then
        lH := lTexture3D.Dim[2] //height anterior-posterion
      else
        lH := lTexture3D.Dim[3]; //sagittal width head-foot;
      if lH > lMaxHt then
        lMaxHt := lH;
      if lX < lCols then
        lWx := round(lWx * (1-abs(lMosaic.HOverlap)));
      lWid := lWid + lWx;
    end; //for each Col
    if lWid > lMaxWid then
      lMaxWid := lWid;
    //lMosaic.VOverlap := 0.5;
    if lY < lRows then
      lT := lT + round(lMaxHt * (1-abs(lMosaic.VOverlap)))
    else
      lT := lT + lMaxHt;
  end;//for each row

(*  for lY := 1 to lRows do
  for lX := 1 to lCols do
      if lMosaic.Text[lX,lY] then
      GLForm1.caption := 'xxx' +'  '+inttostr(random(222));*)
  lHtx := lT;
  lWidx := lMaxWid;
  if noDraw then begin
     //GLForm1.Caption :=':> '+ inttostr(random(888));
    exit;

  end;
  (*for lY := 1 to lRows do
  for lX := 1 to lCols do
      if lMosaic.Text[lX,lY] then
      GLForm1.caption := 'xxx' +'  '+inttostr(random(222));
         *)
  go2D;
  glPushAttrib (GL_ENABLE_BIT);
  glDisable (GL_LIGHTING);
  glBlendFunc (GL_ONE,GL_ZERO{GL_DST_COLOR, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA});
  glEnable (GL_TEXTURE_3D);
  //
  //second pass - draw the images...
  //fx(lT ,lMaxHt);
  //glc:=TGLCanvas.Create(lWidx, lHtx);


  lMinMM := nan;
  if lMosaic.VOverlap < 0 then begin
    lY := lRows;
    lT := lT - lMaxHt;//round(lMaxHt * (1-abs(lMosaic.VOverlap)));
  end else begin
    lY := 1;
    lT := 0;
  end;

  repeat
    lMaxHt := 0;
    if lMosaic.HOverlap < 0 then begin
      lX := lCols;
      lWid := lMaxWid;
    end else begin
      lX := 1;
      lWid := 0;
    end;
    repeat
    //for lX := 1 to lCols do begin
      //determine width
      if (lMosaic.Orient[lX,lY] = kEmptyOrient) then
        lWx := 0
      else if (lMosaic.Orient[lX,lY] = kAxialOrient) or (lMosaic.Orient[lX,lY] =kCoronalOrient) then
        lWx := lTexture3D.Dim[1] //width left-right
      else
        lWx := lTexture3D.Dim[2]; //sagittal width anterior-posterion;
      //compute height
      if (lMosaic.Orient[lX,lY] = kEmptyOrient) then
        lH := 0
      else if (lMosaic.Orient[lX,lY] = kAxialOrient)  then
        lH := lTexture3D.Dim[2] //height anterior-posterion
      else
        lH := lTexture3D.Dim[3]; //sagittal width head-foot;
      if lH > lMaxHt then
        lMaxHt := lH;
      if (lMosaic.HOverlap < 0) and (lX = lCOls) then
        lOffset := lWx
      else
        lOffset := round(lWx * (1-abs(lMosaic.HOverlap)));
      if lMosaic.HOverlap < 0 then
        lWid := lWid - lOffset;

      DrawSlice(lMosaic.Slices[lX,lY], lWid,lT,lScrnW,lScrnH,lMosaic.Orient[lX,lY], lTexture3D);
      lSliceMM := SliceMM (lMosaic.Slices[lX,lY], lMosaic.Orient[lX,lY]);
      if lMinMM = NaN then begin
        lMinMM := lSliceMM;
        lMaxMM := lSliceMM;
      end;
      if lSliceMM < lMinMM then
        lMinMM := lSliceMM;
      if lSliceMM > lMaxMM then
        lMaxMM := lSLiceMM;
      //DrawText2(lMosaic.Slices[lX,lY], lWid,lT,rci);
      if lMosaic.HOverlap < 0 then begin
        dec(lX);
      end else begin
        inc(lX);
        lWid := lWid + lOffset;
      end;
    until (lX < 1) or (lX > lCols);
    if lMosaic.VOverlap < 0 then
      lT := lT - round(lMaxHt * (1-abs(lMosaic.VOverlap)))
    else
      lT := lT + round(lMaxHt * (1-abs(lMosaic.VOverlap)));
    if lMosaic.VOverlap < 0 then
      dec(lY)
    else
      inc(lY);
  until (lY < 1) or (lY > lRows);
  //Finish drawing screen
(*    glDisable(GL_TEXTURE_3D);
  glPopAttrib;
  end2D;
  *)
  //third pass - draw the text
 //compute range of slices, to determine how many decimals to display...
 lSliceMM := lMaxMM - lMinMM;
 if lSliceMM > 10 then
  lDec := 0
 else if lSliceMM > 1 then
  lDec := 1
 else
  lDec := 2;
//for OSX we need to run these prior to text...
      glDisable(GL_TEXTURE_3D);
  glPopAttrib;
  end2D;

 glc:=TGLCanvas.Create(lWidx, lHtx);
//DrawtextX (10,10, 'xxx', rci,true);
  //fx(lHtx ,lMaxHt);
  if lMosaic.VOverlap < 0 then begin
    lY := lRows;
    lT := lHtx - lMaxHt;//round(lMaxHt * (1-abs(lMosaic.VOverlap)));
    //lT := lHtx;//round(lMaxHt * (1-abs(lMosaic.VOverlap)));
  end else begin
    lY := 1;
    lT := 0;
  end;
  repeat
    lMaxHt := 0;
    if lMosaic.HOverlap < 0 then begin
      lX := lCols;
      lWid := lMaxWid;
    end else begin
      lX := 1;
      lWid := 0;
    end;
    repeat
    //for lX := 1 to lCols do begin
      //determine width
      if (lMosaic.Orient[lX,lY] = kEmptyOrient) then
        lWx := 0
      else if (lMosaic.Orient[lX,lY] = kAxialOrient) or (lMosaic.Orient[lX,lY] =kCoronalOrient) then
        lWx := lTexture3D.Dim[1] //width left-right
      else
        lWx := lTexture3D.Dim[2]; //sagittal width anterior-posterion;
      //compute height
      if (lMosaic.Orient[lX,lY] = kEmptyOrient) then
        lH := 0
      else if (lMosaic.Orient[lX,lY] = kAxialOrient)  then
        lH := lTexture3D.Dim[2] //height anterior-posterion
      else
        lH := lTexture3D.Dim[3]; //sagittal width head-foot;
      if lH > lMaxHt then
        lMaxHt := lH;
      if (lMosaic.HOverlap < 0) and (lX = lCOls) then
        lOffset := lWx
      else
        lOffset := round(lWx * (1-abs(lMosaic.HOverlap)));
      if lMosaic.HOverlap < 0 then
        lWid := lWid - lOffset;
      //DrawSlice(lMosaic.Slices[lX,lY], lWid,lT,lScrnW,lScrnH,lMosaic.Orient[lX,lY], lTexture3D);
      //if (lY <> 1) and (lX = 1) then
      if lMosaic.Text[lX,lY]  then begin
//        MosaicPrefsForm.caption := inttostr(lX)+' '+inttostr(lY)+'  '+inttostr(random(888));
//    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//    glEnable (GL_BLEND);
//    GLForm1.GLWindowsBitmapFont1.TextOut(rci, 20, 20, 'x', $FFFFFF);

        DrawText(glc,lMosaic.Slices[lX,lY],round((lWid+(lWx div 2))), lT,lMosaic.Orient[lX,lY],lDec,rci);
      end;
      if lMosaic.HOverlap < 0 then begin
        dec(lX);
      end else begin
        inc(lX);
        lWid := lWid + lOffset;
      end;
    until (lX < 1) or (lX > lCols);
    if lMosaic.VOverlap < 0 then
      lT := lT - round(lMaxHt * (1-abs(lMosaic.VOverlap)))
    else
      lT := lT + round(lMaxHt * (1-abs(lMosaic.VOverlap)));
    if lMosaic.VOverlap < 0 then
      dec(lY)
    else
      inc(lY);
  until (lY < 1) or (lY > lRows);  
  glc.free;
  //GLForm1.caption :=inttostr(random(888))+' :(';

  //    DrawText(glc,lMosaic.Slices[lX,lY],round((lWid+(lWx div 2))/ lScrnW),round(lT/lScrnH),rci);
(*Windows...?
    glDisable(GL_TEXTURE_3D);
  glPopAttrib;
  end2D;
*)

end;

end.
