Unit texture_3d_unit;
{$include options.inc}
Interface
{$H+}
{$DEFINE dontLOADDUMMY}   //<-use LOADDUMMY to test with dynamically generated volume

Uses
{$IFNDEF FPC}
  gziod,
{$ELSE}
  gzio2,
{$ENDIF}
{$IFDEF Unix} {$ELSE} Windows,{$ENDIF}
  SysUtils,
  Dialogs,
  OpenGL1x,GraphicsMathLibrary,
  ExtCtrls, ComCtrls, nifti_hdr, define_types;
Type
  TTexture =  RECORD //peristimulus plot
    NIFTIhdr: TNIFTIHdr;
    PixMM,Scale: array[1..3] of single;
    Dim,FiltDim : array [1..3] of integer;
    DataType,BytesPerVoxel : integer;
    EdgeBias,EdgeGain: byte;
    MRA: boolean; //use maximum intensity projection for angiography....
    WindowScaledMax,WindowScaledMin,MaxThreshScaled,MinThreshScaled: single;
    WindowHisto,UnscaledHisto: HistoRA;
    RawUnscaledImg8,RawUnscaledImgRGBA,FiltImg,EdgeImg,OverlayImg: bytep0;
    RawUnscaledImg16: SmallIntP0;
    RawUnscaledImg32: SingleP0;
  end;
  procedure Float64ToFloat32 (var lHdr: TMRIcroHdr; var lImgBuffer: byteP);
 // procedure Int32ToFloat (var lHdr: TMRIcroHdr; var lImgBuffer: byteP);
function NIFTIhdr_LoadImg (var lFilename: string; var lHdr: TMRIcroHdr; var lImgBuffer: byteP; lVolume: integer): boolean;
Function Load_From_NIfTI (var lTexture: TTexture; Const F_FileName : String; lPowerOfTwo: boolean; lVol: integer) : boolean;
Procedure InitTexture (var lTexture: TTexture);
Procedure SetLengthB (var lPtr: Bytep0;lBytes: integer);
function NIFTIvolumes (var lFilename: string): integer;
procedure EdgeBiasGain (lBiasIn,lGainIn: integer;var  lLUT: TLUTb);
Implementation
uses
{$IFDEF ENABLECLUT}clutprefs,{$ENDIF}
nii_reslice,reorient,histogram, mainunit{gprefs};

procedure EdgeBiasGain (lBiasIn,lGainIn: integer;var  lLUT: TLUTb);
//input values 0..100
{http://dept-info.labri.fr/~schlick/DOC/gem2.html
http://dept-info.labri.fr/~schlick/publi.html
Fast Alternatives to Perlin's Bias and Gain Functions
Christophe Schlick Graphics Gems IV, p379-382, April 1994  }
var
	lIndex: integer;
	lA,lT,lBias,lG,lGain,lResult: single;
begin
  lA := 100-lBiasIn;
	lA := lA/100;
  if lA > 1 then
    lA := 1;
  if lA = 0 then
    lA := 0.000001;
  lG := 100-lGainIn;
	lG := (lG)/100;
        if lG = 0 then
           lG := 0.00001;
        if lG = 1 then
           lG := 0.99999;
	 for lIndex := 0 to (255) do begin
		 lT := lIndex/255;
                 //apply bias
		 lT := (lt/((1/la-2)*(1-lt)+1)) ;
                 //next apply gain
                 if lT < 0.5 then
                      lGain := (lT/((1/lG-2)*(1-2*lT)+1))
                 else
                     lGain := (( (1/lG-2)*(1-2*lT)-lT ) / ( (1/lG-2)*(1-2*lT)-1 ) );
                 lGain := lGain / lT;
                 lResult := (255*lT*lGain);
                 if lResult > 255 then
                    lResult := 255;
                 if lResult < 0 then
                    lResult := 0;
		 lLUT[lIndex] := round(lResult);
		 //lLUT[lIndex] := 255-lLUT[lIndex];

	 end;
end;


function NIFTIvolumes (var lFilename: string): integer;
var lHdr: TMRIcroHdr;
begin
  result := -1;
    if not NIFTIhdr_LoadHdr (lFilename, lHdr) then
      exit;
  result := lHdr.NIFTIhdr.dim[4];
end;



Procedure SetLengthB (var lPtr: Bytep0;lBytes: integer);
begin
    if lPtr <> nil then
      freemem(lPtr);
    if lBytes < 1 then begin
      lPtr := nil;
      exit;
    end;
    getmem(lPtr,lBytes);
end;

Procedure SetLength16 (var lPtr: SmallIntP0;lBytes: integer);
begin
    if lPtr <> nil then
      freemem(lPtr);
    if lBytes < 1 then begin
      lPtr := nil;
      exit;
    end;
    getmem(lPtr,lBytes);
end;

Procedure SetLength32 (var lPtr: SingleP0;lBytes: integer);
begin
    if lPtr <> nil then
      freemem(lPtr);
    if lBytes < 1 then begin
      lPtr := nil;
      exit;
    end;
    getmem(lPtr,lBytes);
end;

Procedure InitTexture (var lTexture: TTexture);
var
  lI: integer;
begin
  with lTexture do begin
    for lI := 1 to 3 do begin
      PixMM[lI] := 0;
      Scale[lI] := 1;
      Dim[lI] := 0;
      FiltDim[lI] := 0;
    end;
    NIFTIhdr.scl_slope := 1;
    NIFTIhdr.scl_inter := 0;
    EdgeBias := 127;
    EdgeGain := 127;
    MRA := false;
    SetLengthB(FiltImg,0);
    SetLengthB(EdgeImg,0);
    SetLengthB(OverlayImg,0);

    SetLengthB(RawUnscaledImg8,0);
    SetLength16(RawUnscaledImg16,0);
    SetLength32(RawUnscaledImg32,0);
    DataType := 0;
    BytesPerVoxel := 0;
  end; //with lTexture
end; //Init

procedure Data_Type_To_Texel_Byte_Size (var lTexture: TTexture);
//determine bytes per voxel
Begin
 with lTexture do begin
  Case lTexture.DataType Of
    GL_COLOR_INDEX : BytesPerVoxel := 1;
    GL_STENCIL_INDEX : BytesPerVoxel := 1;
    GL_DEPTH_COMPONENT :BytesPerVoxel := 1;
    GL_RED : BytesPerVoxel := 1;
    GL_GREEN : BytesPerVoxel := 1;
    GL_BLUE : BytesPerVoxel := 1;
    GL_ALPHA : BytesPerVoxel := 1;
    GL_RGB : BytesPerVoxel := 3;
    GL_RGBA : BytesPerVoxel := 4;
    GL_LUMINANCE : BytesPerVoxel := 1;
    GL_LUMINANCE_ALPHA : BytesPerVoxel := 2;
    Else BytesPerVoxel := 4;
  End; //Case
 end;//with lTexture
End; //Data_Type_To_Texel_Byte_Size

Procedure NormalizeVector (var lX,lY,lZ: single);
var
  V: single;
begin
  V := Sqrt(sqr(lx)+sqr(ly)+sqr(lZ));
  if V = 0 then
    exit;
  lX := lX/V;
  lY := lY/V;
  lZ := lZ/V;
end;

function IsMRA (var lHdr: TNIfTIHdr): boolean;
//return TRUE if first 3 characters of intention string are MRA - helps detect images that require MIP
begin
  result := false;
  if length (lHdr.intent_name) < 3 then
    exit;
  if upcase(lHdr.intent_name[1]) <> 'M' then
    exit;
  if upcase(lHdr.intent_name[2]) <> 'R' then
    exit;
  if upcase(lHdr.intent_name[3]) <> 'A' then
    exit;
  result := true;
end;

{$IFDEF LOADDUMMY}
//generate volume dynamically instead of loading from disk
function NIFTIhdr_LoadDummyImg (var lFilename: string; var lHdr: TMRIcroHdr; var lImgBuffer: byteP): boolean;
const
  kSz = 128;
var
  lX,lY,lZ,lFileBytes,lI: integer;
begin
  NIFTIhdr_ClearHdr(lHdr);
  with lHdr.NIFTIhdr do begin
    dim[1] := kSz;
    dim[2] := kSz;
    dim[3] := kSz;
    bitpix := 8;//vc16; {8bits per pixel, e.g. unsigned char 136}
    DataType := 2;//vc4;{2=unsigned char, 4=16bit int 136}
  end;
  lFileBytes := kSz * kSz * kSz;
  GetMem(lImgBuffer,lFileBytes);
  for lI := 1 to lFilebytes do
    lImgBuffer^[lI] := 0;
  lI := 0;
  for lZ := 1 to kSz do begin
    for lY := 1 to kSz do begin
      for lX := 1 to kSz do begin
        inc(lI);
        if (lZ > 10) and (lZ < 20) then
          lImgBuffer^[lI] := lX;
        if (lY > 10) and (lY < 20) then
          lImgBuffer^[lI] := lX;

      end;
    end;
  end;
  result := true;
end;
{$ENDIF}



function NIFTIhdr_LoadImg (var lFilename: string; var lHdr: TMRIcroHdr; var lImgBuffer: byteP; lVolume: integer): boolean;
//loads img to byteP - if this returns successfully you must freemem(lImgBuffer)
var
   lExt,lImgName: string;
   lVolOffset,lnVol,lVol,lFileBytes,lImgBytes: integer;
   lBuf: ByteP;
   lGZin: boolean;
   lInF: File;
begin
    result := false;
    if not NIFTIhdr_LoadHdr (lFilename, lHdr) then begin
        //Showmessage('LoadImg: Unable to read as NifTI/Analyze' + lFilename);
        exit;
    end;
    lExt := UpCaseExt(lFilename);
    lGZin := ExtGZ(lFilename);
    if lExt = '.VOI' then
       lGZin := true;
   if lHdr.NIFTIhdr.dim[4] < 1 then
    lHdr.NIFTIhdr.dim[4] := 1;
   if lHdr.NIFTIhdr.dim[5] < 1 then
    lHdr.NIFTIhdr.dim[5] := 1;
   lnVol := lHdr.NIFTIhdr.dim[4]*lHdr.NIFTIhdr.dim[5]; //Time+Direction
   lVol := lVolume;
   if (lVol < 1) or (lVol > lnVol) then
    lVol := lnVol;
   if lHdr.NIFTIhdr.datatype = kDT_RGB then begin
      lHdr.NIFTIhdr.bitpix := 24;
      lVol := 1; //read all RGB planes, later on we can separate different planes
   end;
   lImgBytes := lHdr.NIFTIhdr.dim[1]*lHdr.NIFTIhdr.dim[2]*lHdr.NIFTIhdr.dim[3]*(lHdr.NIFTIhdr.bitpix div 8);
   lVolOffset := (lVol-1) * lImgBytes;
   if lExt ='.HDR' then
       lImgName :=  changefileext(lFilename,'.img')
   else
       lImgName := lFilename;
   if not fileexists(lImgName) then begin
       Showmessage('LoadImg Error: Unable to find '+lImgName);
       exit;
   end;
   if (not lGZin) and (FSize (lImgName) < ( lImgBytes)) then begin
       Showmessage('LoadImg Error: File smaller than expected  '+lImgName);
       exit;
   end;
   lFileBytes := lImgBytes;
   GetMem(lImgBuffer,lFileBytes);
   Filemode := 0;  //Read Only - allows us to open images where we do not have permission to modify
   if lGZin then begin
      lBuf := @lImgBuffer^[1];
      UnGZip (lImgName,lBuf, lVolOffset+round(lHdr.NIFtiHdr.vox_offset),lImgBytes);
   end else begin
       AssignFile(lInF, lImgName);
       Reset(lInF,1);
       Seek(lInF,lVolOffset+round(lHdr.NiftiHdr.vox_offset));
       BlockRead(lInF, lImgBuffer^[1],lImgBytes);
       CloseFile(lInF);
   end;
   Filemode := 2;  //Read/Write
   result := true;
end;

procedure PowerOfTwo(Var lDim, lPad: integer);
//return 2^n that is equal or larger than the input lDim value
// e.g. if lDim is 129..256 then this procedure will return 256
var
  lResult: integer;
begin
  lResult := 2;
  while (lResult < lDim) do
    lResult := lResult * 2;
  lPad := (lResult - lDim) div 2;
  lDim := lResult;
end;

procedure PadPrime(Var lDim : integer);
//return 2^n that is equal or larger than the input lDim value
// e.g. if lDim is 129..256 then this procedure will return 256
const
  knPrime = 97;
  kPrime : array  [1..knPrime] of integer =
      (5,7,11,13,17,19,23,29,31,37,41,43,47,53,59,61,67,71,73,79,83,89,97,
      101,103,107,109,113,127,131,137,139,149,151,157,163,167,173,179,181,191,
      193,197,199,211,223,227,229,233,239,241,251,257,263,269,271,277,281,283,
      293,307,311,313,317,331,337,347,349,353,359,367,373,379,383,389,397,401,
      409,419,421,431,433,439,443,449,457,461,463,467,479,487,491,499,503,509,521,523);

var
  lResult,lI: integer;
begin
  //lPad := 0;
  lResult := lDim;
  if lDim < 4 then
    lResult := 4
  else begin
    for lI := 1 to knPrime do begin
      if lDim = kPrime[lI] then begin
        lResult := lDim + 1;
      end;
    end;
  end;
  lDim := lResult;
end;

procedure PadThin(Var lDim : integer);
//We will strip the top slice - this is a problem if the image is very thin in the Z dimension
begin
  if lDim < 4 then
    inc(lDim);
end;


procedure ClearSlice (var lTexture: TTexture; lSlice: integer; lVal: single);
var
  lSliceVox, lOffset, lVox,lVali: integer;
begin
  if (lSlice < 1) or (lSlice > lTexture.Dim[3]) then
    exit;
  lSliceVox := lTexture.Dim[1] * lTexture.Dim[2];
  if lSliceVox < 1 then
    exit;
  lOffset := (lSlice-1) * lSliceVox;
  lVali := round(lVal);
  if lTexture.RawUnscaledImg8 <> nil then
     for lVox := 0 to (lSliceVox-1) do //set padded regions to minimum value
        lTexture.RawUnscaledImg8^[lVox+lOffset] := lVali
  else if lTexture.RawUnscaledImg16 <> nil then
     for lVox := 0 to (lSliceVox-1) do //set padded regions to minimum value
        lTexture.RawUnscaledImg16^[lVox+lOffset] := lVali
  else if lTexture.RawUnscaledImg32 <> nil then
     for lVox := 0 to (lSliceVox-1) do //set padded regions to minimum value
        lTexture.RawUnscaledImg32^[lVox+lOffset] := lVal
  else if lTexture.RawUnscaledImgRGBA <> nil then begin
     lOffset := ((lSlice-1) * lSliceVox * 4) +1;//GREEN USED FOR ALPHA
     for lVox := 0 to (lSliceVox-1) do begin //set padded regions to minimum value
        lTexture.RawUnscaledImgRGBA^[lOffset] := 0;  //
        inc(lOffset,4);
     end;
  end;
end;

procedure Float64ToFloat32 (var lHdr: TMRIcroHdr; var lImgBuffer: byteP);
var
  lI,lInVox: integer;
  l64Buf : DoubleP;
  lV: double;
  l32TempBuf,l32Buf : SingleP;
begin
	  if lHdr.NIFTIHdr.datatype <> kDT_DOUBLE then
      exit;
    lInVox :=  lHdr.NIFTIhdr.dim[1] *  lHdr.NIFTIhdr.dim[2] * lHdr.NIFTIhdr.dim[3];
    l64Buf := DoubleP(lImgBuffer );
    GetMem(l32TempBuf ,lInVox*sizeof(single));
    if not lHdr.DiskDataNativeEndian then begin
        for lI := 1 to lInVox do begin
                         try
                            l32TempBuf^[lI] := Swap64r(l64Buf^[lI])
                         except
                            l32TempBuf^[lI] := 0;
                         end; //except
        end; //for
    end else begin  //convert integer to float
			 for lI := 1 to lInVox do begin
                        // try
                        lV := l64Buf^[lI];
                            l32TempBuf^[lI] := lV;//
                       //  except
                       //     l32TempBuf^[lI] := 0;
                       //  end; //except
       end;
    end;
    freemem(lImgBuffer);
    GetMem(lImgBuffer ,lInVox*sizeof(single));
    l32Buf := SingleP(lImgBuffer );
    Move(l32TempBuf^,l32Buf^,lInVox*sizeof(single));
    freemem(l32TempBuf);
    for lI := 1 to lInVox do
			if specialsingle(l32Buf^[lI]) then l32Buf^[lI] := 0.0;
    lHdr.NIFTIHdr.datatype := kDT_FLOAT;
    lHdr.DiskDataNativeEndian := true;
    lHdr.NIFTIhdr.bitpix := 32;
end;//Float64ToFloat32

(*procedure Float32RemoveNAN (var lHdr: TMRIcroHdr; var lImgBuffer: byteP);
//set "Not-A-Number" values to be zero... SPM uses NaN for voxels it can not compute
var
  lI,lInVox: integer;
  l32Buf : SingleP;
begin
	  if lHdr.NIFTIHdr.datatype <> kDT_FLOAT then
      exit;
    lInVox :=  lHdr.NIFTIhdr.dim[1] *  lHdr.NIFTIhdr.dim[2] * lHdr.NIFTIhdr.dim[3];
    l32Buf := SingleP(lImgBuffer );
    for lI := 1 to lInVox do
			if specialsingle(l32Buf^[lI]) then l32Buf^[lI] := 0.0;
end;//Float32RemoveNAN
*)

procedure PadTransform(var lPadX,lPadY,lPadZ: integer; var lHdr : TNIFTIHdr);
var
     lInMat,lTransMat,lResidualMat: TMatrix;
begin
  if (lPadX=0) and (lPadY=0) and (lPadZ = 0) then
    exit;
   lInMat := Matrix3D (
    lHdr.srow_x[0],lHdr.srow_x[1],lHdr.srow_x[2],lHdr.srow_x[3],
    lHdr.srow_y[0],lHdr.srow_y[1],lHdr.srow_y[2],lHdr.srow_y[3],
    lHdr.srow_z[0],lHdr.srow_z[1],lHdr.srow_z[2],lHdr.srow_z[3],
    0,0,0,1);
   lTransMat := Matrix3D (
    1,0,0,-lPadX,
    0,1,0,-lPadY,
    0,0,1,-lPadZ,
    0,0,0,1);
  lResidualMat := multiplymatrices(lInMat,lTransMat); //source
  lHdr.srow_x[0] := lResidualMat.Matrix[1,1];
  lHdr.srow_x[1] := lResidualMat.Matrix[1,2];
  lHdr.srow_x[2] := lResidualMat.Matrix[1,3];
  lHdr.srow_y[0] := lResidualMat.Matrix[2,1];
  lHdr.srow_y[1] := lResidualMat.Matrix[2,2];
  lHdr.srow_y[2] := lResidualMat.Matrix[2,3];
  lHdr.srow_z[0] := lResidualMat.Matrix[3,1];
  lHdr.srow_z[1] := lResidualMat.Matrix[3,2];
  lHdr.srow_z[2] := lResidualMat.Matrix[3,3];
  lHdr.srow_x[3] := lResidualMat.Matrix[1,4];
  lHdr.srow_y[3] := lResidualMat.Matrix[2,4];
  lHdr.srow_z[3] := lResidualMat.Matrix[3,4];
end;

Function Load_From_NIfTI (var lTexture: TTexture; Const F_FileName : String; lPowerOfTwo: boolean; lVol: integer) : boolean;
Var
  lCLUT: TLUT;
  lPadX,lPadY,lPadZ,ValueI,lI,lZ,lY,lX,lSLiceStart,lLineStart,
  lInSliceSz,lPos,lInVox,lOutVox,lIndex: integer;
  lMinS,lMaxS,lValueS,lIntercept,lSlope: single;
  lMaxB,lMinB,lR,lG,lB,ValueB: byte;
  lFilename: string;
  lHdr: TMRIcroHdr;
  lImgBuffer: byteP;
  l16Buf : SmallIntP;
  l32Buf : SingleP;
Begin //Proc Load_From_NIfTI
//There are two approaches to spatial coordinates -
//  1.) Multiply NIfTI matrix by OpenGL coordinates: simple
//      but angle of things like clip planes is not intuitive
//  2.) Reorient data to be in closest orthogonal plane to OpenGL space - see reorientcore
//  Here I take the latter approach
    result :=false;
    lFilename := F_Filename;
    InitTexture(lTexture);
    {$IFDEF LOADDUMMY}
    if not NIFTIhdr_LoadDummyImg (lFilename, lHdr, lImgBuffer) then
      exit;
    {$ELSE}
    if not NIFTIhdr_LoadImg (lFilename, lHdr, lImgBuffer,lVol) then
      exit;
    {$ENDIF}
    lTexture.NIFTIhdr := lHdr.NIFTIhdr;
    lTexture.MRA := IsMRA(lHdr.NIFTIHdr);
    //lTexture.ScaleSlope := lHdr.NIFTIhdr.scl_slope;
    //if lTexture.ScaleSlope = 0 then
    //  lTexture.ScaleSlope := 1;
    //lTexture.ScaleIntercept := lHdr.NIFTIhdr.scl_inter;
    //Float32Endian (lHdr,lImgBuffer);
    Int32ToFloat(lHdr,lImgBuffer);
    Float64ToFloat32(lHdr,lImgBuffer);
    NIFTIhdr_UnswapImg (lHdr,lImgBuffer); //ensures image data is in native byteorder
    Float32RemoveNAN(lHdr,lImgBuffer);
    NIFTIhdr_MinMaxImg (lHdr,lImgBuffer); //sets global minmax
    if not (lHdr.NIFTIHdr.datatype in [kDT_UNSIGNED_CHAR, kDT_SIGNED_SHORT, kDT_FLOAT,kDT_RGB]) then begin
      freemem(lImgBuffer);
      showmessage('Error: currently only able to read 8 or 16-bit integers, 32-bit floats, or 24-bit RGB data.');
      exit;//abort - unsupported format
    end;

    ReorientCore(lHdr.NIFTIHdr, lImgBuffer);
    lInVox :=  lHdr.NIFTIhdr.dim[1] *  lHdr.NIFTIhdr.dim[2] * lHdr.NIFTIhdr.dim[3];
    //nifti_normal(lHdr);
    //XYZtoGLCoordinates (0,0,1,M_Up_X,M_Up_Y,lTexture.Up[3]);
    //Up is the vector [0,0,1]*Rotation matrix
    //As the zeros have no influence, it is faster to read the Z-row of the matrix
    //set the right position
    //lTexture.Right[1] := lHdr.NIFTIhdr.srow_x[0];//X value
    //lTexture.Right[2] := lHdr.NIFTIhdr.srow_x[1];//y
    //lTexture.Right[3] := lHdr.NIFTIhdr.srow_x[2];
    //If this is not set, assume an identity matrix, with the third row being 0,0,1
    //if (lTexture.Right[1] = lTexture.Right[2]) and (lTexture.Right[1] = lTexture.Right[3]) and (lTexture.Right[1] = 0) then
    //  lTexture.Right[1] := 1;
(*    NormalizeVector(lTexture.Right[1],lTexture.Right[2],lTexture.Right[3]);
    //set the anterior direction. You could figure this out from
    lTexture.Front[1] := lHdr.NIFTIhdr.srow_y[0];//X value
    lTexture.Front[2] := lHdr.NIFTIhdr.srow_y[1];//y
    lTexture.Front[3] := lHdr.NIFTIhdr.srow_y[2];
    //If this is not set, assume an identity matrix, with the third row being 0,0,1
    if (lTexture.Front[1] = lTexture.Front[2]) and (lTexture.Front[1] = lTexture.Front[3]) and (lTexture.Front[1] = 0) then
      lTexture.Front[2] := 1;
    NormalizeVector(lTexture.Front[1],lTexture.Front[2],lTexture.Front[3]);
    //Set the Up [Superior] direction
    lTexture.Up[1] := lHdr.NIFTIhdr.srow_z[0];//X value
    lTexture.Up[2] := lHdr.NIFTIhdr.srow_z[1];//y
    lTexture.Up[3] := lHdr.NIFTIhdr.srow_z[2];
    //If this is not set, assume an identity matrix, with the third row being 0,0,1
    if (lTexture.Up[1] = lTexture.Up[2]) and (lTexture.Up[1] = lTexture.Up[3]) and (lTexture.Up[1] = 0) then
      lTexture.Up[3] := 1;
    //NormalizeVector(lTexture.Up[1],lTexture.Up[2],lTexture.Up[3]);
   *)
    for lI := 1 to 3 do
      lTexture.Dim[lI] := lHdr.NIFTIhdr.dim[lI];
    //next get pixel spacing in mm...
    lTexture.PixMM[1] := abs(lHdr.NIFTIhdr.PixDim[1]);//use absolute -some people use negative values for flipped dimensions
    lTexture.PixMM[2] := abs(lHdr.NIFTIhdr.PixDim[2]);
    lTexture.PixMM[3] := abs(lHdr.NIFTIhdr.PixDim[3]);
    if lTexture.PixMM[1] = 0 then
      lTexture.PixMM[1] := 1;
    if lTexture.PixMM[2] = 0 then
      lTexture.PixMM[2] := 1;
    if lTexture.PixMM[3] = 0 then
      lTexture.PixMM[3] := 1;
    //next: as of 2008, OpenGL crashes if slice dimension is a prime number
    //  a quick an dirty solution is to pad images...
    //  this slows down the writing of our output images
    //  Modern NVidia cards do not care if the number of slices [Z-dimension] are even, but ATI X1600 Radeons does
    lPadX := 0;
    lPadY := 0;
    lPadZ := 0;
    PadThin(lTexture.Dim[3]);
    if (lPowerOfTwo) or (gPrefs.ForcePowerOfTwo) then begin
        PowerOfTwo(lTexture.Dim[1],lPadX);
        PowerOfTwo(lTexture.Dim[2],lPadY);
        PowerOfTwo(lTexture.Dim[3],lPadZ);
        PadTransform(lPadX,lPadY,lPadZ, lHdr.NIFTIHdr);
    {$IFDEF UNIX}
    end else  begin
      //There is a real problem in padding data, as any overlays/ROIs/VOIs need to be unclipped
      //otherwise FSL/SPM will not load them (as spatial dimensions will differ from background
      //..always add slice to top - this gives us a free clip plane
        PadPrime(lTexture.Dim[1] {,lPadX});
        PadPrime(lTexture.Dim[2]{,lPadY});
        //if gPrefs.ForcePowerOfTwoSlices then
        //   PowerOfTwo(lTexture.Dim[3],lPadZ);
      //PowerOfTwo(lTexture.Dim[3],lPadZ);
    {$ELSE}
    end else begin
      //Windows OpenGL implementation does not require padding data..
    {$ENDIF}
    end;//not power of two
    //Overlays need to be aligned - the next 3 lines adjust the background relative to the overlay
    {lPadX := 0; //not sure if this is correct
    lPadY := 0; //this has been checked...
    lPadZ := 0;//not sure if this is correct
    }
    lTexture.DataType := GL_RGBA ;
    Data_Type_To_Texel_Byte_Size (lTexture);
    lOutVox :=  lTexture.Dim[1] * lTexture.Dim[2] * lTexture.Dim[3];
    if lHdr.NIFTIHdr.datatype = kDT_UNSIGNED_CHAR then begin //Luminance
      SetLengthB(lTexture.RawUnscaledImg8,   lOutVox);
      if lOutVox = lInVox then begin //DEC09... no padding
        Move(lImgBuffer^,lTexture.RawUnscaledImg8^,lInVox);//source/dest
      end else begin
        lI := lImgBuffer^[1];
        for lZ := 0 to (lOutVox-1) do //set padded regions to minimum value
          lTexture.RawUnscaledImg8^[lZ] := lI;
        lPos := 0;
        for lZ := 0 to (lHdr.NIFTIhdr.dim[3]-1) do begin
          lSliceStart := (lPadZ+lZ) * (lTexture.Dim[1]*lTexture.Dim[2]);
          for lY := 0 to (lHdr.NIFTIhdr.dim[2]-1) do begin
            lLineStart := lSLiceStart+((lPadY+lY) * lTexture.Dim[1]);
            lI := lPadX;
            for lX := 0 to (lHdr.NIFTIhdr.dim[1]-1) do begin
              inc(lPos);
              lTexture.RawUnscaledImg8^[lLineStart+lI ] := lImgBuffer^[lPos];
              inc(lI);
            end;//for X
          end;//for Y
        end;//for Z
      end;//padded
    end; //8bit
    if lHdr.NIFTIHdr.datatype = kDT_SIGNED_SHORT then begin //16-bit
      l16Buf := SmallIntP(lImgBuffer);
      {UnswapImg already computed...
      if not lHdr.DiskDataNativeEndian then
        for lPos := 1 to lInVox do
			    l16Buf^[lPos] := Swap(l16Buf^[lPos]);}
      SetLength16(lTexture.RawUnscaledImg16,   lOutVox* 2);
      if lOutVox = lInVox then begin //no padding
        Move(l16Buf^,lTexture.RawUnscaledImg16^,lOutVox* 2);//source/dest
      end else begin //not padded
        lI := l16Buf^[1];
        for lZ := 0 to (lOutVox-1) do //set padded regions to minimum value
          lTexture.RawUnscaledImg16^[lZ] := lI;
        lPos := 0;
        for lZ := 0 to (lHdr.NIFTIhdr.dim[3]-1) do begin
          lSliceStart := (lPadZ+lZ) * (lTexture.Dim[1]*lTexture.Dim[2]);
          for lY := 0 to (lHdr.NIFTIhdr.dim[2]-1) do begin
            lLineStart := lSLiceStart+((lPadY+lY) * lTexture.Dim[1]);
            lI := lPadX;
            for lX := 0 to (lHdr.NIFTIhdr.dim[1]-1) do begin
              inc(lPos);
              lTexture.RawUnscaledImg16^[lLineStart+lI ] := l16Buf^[lPos];
              inc(lI);
            end;//for X
          end;//for Y
        end;//for Z
      end;//padded
    end; //16-bit integer datatype
    if lHdr.NIFTIHdr.datatype = kDT_FLOAT then begin //32-bit float
      l32Buf := SingleP(lImgBuffer);
      SetLength32(lTexture.RawUnscaledImg32,   lOutVox* 4);
      if lOutVox = lInVox then begin //no padding
        Move(l32Buf^,lTexture.RawUnscaledImg32^,lOutVox* 4);//source/dest
      end else begin //not padded
        lMinS := l32Buf^[1];
        for lZ := 0 to (lOutVox-1) do //set padded regions to minimum value
          lTexture.RawUnscaledImg32^[lZ] := lMinS;
        lPos := 0;
        for lZ := 0 to (lHdr.NIFTIhdr.dim[3]-1) do begin
          lSliceStart := (lPadZ+lZ) * (lTexture.Dim[1]*lTexture.Dim[2]);
          for lY := 0 to (lHdr.NIFTIhdr.dim[2]-1) do begin
            lLineStart := lSLiceStart+((lPadY+lY) * lTexture.Dim[1]);
            lI := lPadX;
            for lX := 0 to (lHdr.NIFTIhdr.dim[1]-1) do begin
              inc(lPos);
              lTexture.RawUnscaledImg32^[lLineStart+lI ] := l32Buf^[lPos];
              inc(lI);
            end;//for X
          end;//for Y
        end;//for Z
      end;//padded
      //ComputeMinMax( lTexture); //abba
    end; //32-bit FLOAT datatype
    if lHdr.NIFTIHdr.datatype = kDT_RGB then begin //RGB
        SetLengthB(lTexture.RawUnscaledImgRGBA,   lOutVox* lTexture.BytesPerVoxel);
        fillchar(lTexture.RawUnscaledImgRGBA^,lOutVox* lTexture.BytesPerVoxel,0);
      lInSliceSz := lHdr.NIFTIhdr.dim[1]*lHdr.NIFTIhdr.dim[2];
       for lZ := 0 to (lHdr.NIFTIhdr.dim[3]-1) do begin
        lSliceStart := (lPadZ+lZ) * (lTexture.Dim[1]*lTexture.Dim[2]*4);
        lPos := lZ * (lHdr.NIFTIhdr.dim[1]*lHdr.NIFTIhdr.dim[2]*3);//input is planar RGB slices
        for lY := 0 to (lHdr.NIFTIhdr.dim[2]-1) do begin
          lIndex := lSLiceStart+( (lY+lPadY) * lTexture.Dim[1]*4);
          lIndex := lIndex + (lPadX* 4);
          for lX := 0 to (lHdr.NIFTIhdr.dim[1]-1) do begin
            //output stores data as contiguous bytes RGBARGBARGBA...
            //input is planar RRRRRRGGGGGBBBBB
            inc(lPos);
            lR := lImgBuffer^[lPos]; //red plane
            lG := lImgBuffer^[lPos+lInSliceSz]; //green plane
            lB := lImgBuffer^[lPos+lInSliceSz+lInSliceSz]; //blue plane
            valueB := (lR+lG+lB) div 3;
            lTexture.RawUnscaledImgRGBA^[ lIndex] := lR;//R
            lTexture.RawUnscaledImgRGBA^[ lIndex+1] := lG;//G
            lTexture.RawUnscaledImgRGBA^[ lIndex+2] := lB;//B
            lTexture.RawUnscaledImgRGBA^[ lIndex+3] := valueB;
            inc(lIndex,4);
          end;//for Z
        end;//for Y
       end;//for Z
      //end - extra slice to avoid clip bleeding
    end;
    //prepare Output data
    freemem(lImgBuffer);
    //next -generate LUT
    SetLengthB(lTexture.FiltImg,   lOutVox* lTexture.BytesPerVoxel);
    //Dec09 - this next line is slow, perhaps if there is an empty slice???
    //fillchar(lTexture.FiltImg^,lOutVox* lTexture.BytesPerVoxel,0);
    lTexture.FiltDim[1] := lTexture.Dim[1];
    lTexture.FiltDim[2] := lTexture.Dim[2];
    lTexture.FiltDim[3] := lTexture.Dim[3];
    //normalized scale - longest dimension is 1
    lMaxS := FloatMaxVal(lTexture.Dim[1]*lTexture.pixmm[1],lTexture.Dim[2]*lTexture.pixmm[2],lTexture.Dim[3]*lTexture.pixmm[3]);
    if (not gPrefs.ProportionalStretch)  then begin
      lMaxS := FloatMaxVal(lTexture.Dim[1],lTexture.Dim[2],lTexture.Dim[3]);
      lTexture.Scale[1] := lTexture.Dim[1] / lMaxS;
      lTexture.Scale[2] := lTexture.Dim[2] / lMaxS;
      lTexture.Scale[3] := lTexture.Dim[3] / lMaxS;
    end else if (lMaxS = 0) then begin
      lTexture.Scale[1] := 1;
      lTexture.Scale[2] := 1;
      lTexture.Scale[3] := 1;
    end else begin
      lTexture.Scale[1] := (lTexture.Dim[1]*lTexture.PixMM[1]) / lMaxS;
      lTexture.Scale[2] := (lTexture.Dim[2]*lTexture.PixMM[2]) / lMaxS;
      lTexture.Scale[3] := (lTexture.Dim[3]*lTexture.PixMM[3]) / lMaxS;
    end;
    ComputeMinMax(lTexture);
    ClearSlice (lTexture, lTexture.Dim[3],Scaled2Unscaled(lTexture.WindowScaledMin,lTexture) ); //set all voxels in top slice to zero - prevents clipping artifacts
    if  (lHdr.NIFTIHdr.cal_min < lHdr.NIFTIHdr.cal_max) then begin
      //need to create histogram so user will see a CLUT in the 'Color&Transparency' window.... 1/2010
      CreateHistoThresh(lTexture, lTexture.WindowScaledMin, lTexture.WindowScaledMax, lTexture.UnscaledHisto, true,0,lTexture.MinThreshScaled,lTexture.MaxThreshScaled);
      lTexture.MinThreshScaled := lHdr.NIFTIHdr.cal_min;
      lTexture.MaxThreshScaled := lHdr.NIFTIHdr.cal_max;
    end else if lTexture.MRA then
      CreateHistoThresh(lTexture, lTexture.WindowScaledMin, lTexture.WindowScaledMax, lTexture.UnscaledHisto, true,0,lTexture.MinThreshScaled,lTexture.MaxThreshScaled)
    else
      CreateHistoThresh(lTexture, lTexture.WindowScaledMin, lTexture.WindowScaledMax, lTexture.UnscaledHisto, true,0.005,lTexture.MinThreshScaled,lTexture.MaxThreshScaled);
    {$IFDEF ENABLECLUT}

    CLUTForm.DrawHistoX (CLUTForm.GLPolygonInset,lTexture.UnscaledHisto, -0.5,-0.5,1,1);
    CLUTForm.LUTdrop.ItemIndex := 0; //defaults
    CLUTForm.LUTDropChange(nil);
    {$ELSE}
    AutoContrast(gCLUTrec);
    RangeRec(lTexture.MinThreshScaled,lTexture.MaxThreshScaled);
    {$ENDIF}
    lTexture.NIFTIhdr := lHdr.NIFTIhdr;
    for lI := 1 to 3 do
      lTexture.NIFTIhdr.dim[lI] := lTexture.FiltDim[lI];
    result :=true;
End; //Proc Load_From_NIfTI

End.

